"""Defines the global state object for the application.

The global state object is a singleton that holds the current state of the
application. It is used to store the current configuration, Zabbix client,
and other stateful objects.
"""

# This module re-treads all the sins of Harbor CLI's state module.
#
# The lazy-loading, properties with setters/getters, and the singleton pattern
# are all code smells in their own ways (lazy-loading maybe less so in a CLI context).
# The sad thing is that this works well, and it's hard to find a good way to configure
# the application once, save that state, then maintain it for the duration
# of the application's lifetime. The only thing we have to be careful about it
# is to ensure we configure the State object before we use it, and perform
# all imports from other modules inside functions, so that we don't create
# circular imports.
#
# For some reason, I am totally blanking on how to do this in a way that doesn't
# suck when using Typer. With Click, it is easy, because we could just use the
# `@pass_obj` decorator to pass the State object to every command, but Typer
# doesn't have anything like that.
#
# We have a bunch of state we need to access in order to determine things like
# authentication, logging, default values, etc. So having them all gathered
# in this one place is convenient.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from typing import Any
from typing import Optional

from typing_extensions import Self

# This module should not import from other local modules because it's widely
# used throughout the application, and we don't want to create circular imports.
# Runtime imports from other modules should be done inside functions,
# while annotations imports should done in `if TYPE_CHECKING:` blocks.
if TYPE_CHECKING:
    from prompt_toolkit.history import History
    from rich.console import Console

    from zabbix_cli.config.model import Config
    from zabbix_cli.pyzabbix.client import ZabbixAPI

logger = logging.getLogger(__name__)


class State:
    """Application state singleton.

    Object that encapsulates the current state of the application.
    Holds the current configuration, Zabbix client, and other stateful objects.
    """

    _instance = None

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        """Instantiate the singleton object or return the existing instance."""
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    repl: bool = False
    """REPL is active."""

    bulk: bool = False
    """Running in bulk mode."""

    _client: Optional[ZabbixAPI] = None
    """Zabbix API client object."""

    _config: Optional[Config] = None
    """Current Config object (may have overrides)."""

    _config_loaded: bool = False
    """Config has been loaded from file."""

    _config_repl_original: Optional[Config] = None
    """Config object when the REPL was first launched."""

    _console: Optional[Console] = None
    """Stdout Rich console object."""

    _err_console: Optional[Console] = None
    """Stderr Rich console object."""

    token: Optional[str] = None
    """Active Zabbix API auth token."""

    _history: Optional[History] = None

    @property
    def client(self) -> ZabbixAPI:
        """Zabbix API client object.

        Fails if the client is not configured.
        """
        from zabbix_cli.exceptions import ZabbixCLIError

        if self._client is None:
            raise ZabbixCLIError("Not connected to the Zabbix API.")
        return self._client

    @client.setter
    def client(self, client: ZabbixAPI) -> None:
        self._client = client

    @property
    def is_client_loaded(self) -> bool:
        """Zabbix client is configured and ready to use."""
        return self._client is not None

    @property
    def config(self) -> Config:
        """Current Config object."""
        if self._config is None:
            from zabbix_cli.config.model import Config
            from zabbix_cli.logs import configure_logging

            # HACK: configure logging with sample config
            # in order to log the debug message to the default log file
            config = Config.sample_config()
            configure_logging(config.logging)
            logger.debug(
                "Using sample config file as fallback.",
                stacklevel=2,  # See who called this
            )
            return config
        return self._config

    @config.setter
    def config(self, config: Config) -> None:
        """Set the config object and update active configurations.

        Triggers an update of configuration of logging, consoles, etc.
        """
        from zabbix_cli.logs import configure_logging
        from zabbix_cli.output.console import configure_console

        self._config = config
        configure_logging(config.logging)
        configure_console(config)
        # HACK: don't consider using sample config as "loaded".
        # Signals that main callback should create a new config file.
        self._config_loaded = config.sample is False

    @property
    def is_config_loaded(self) -> bool:
        """Config has been loaded from file."""
        return self._config_loaded

    @property
    def console(self) -> Console:
        """Rich console object."""
        # fmt: off
        if not self._console:
            from .output.console import console
            self._console = console
        return self._console
        # fmt: on

    @property
    def err_console(self) -> Console:
        """Rich console object."""
        # fmt: off
        if not self._err_console:
            from .output.console import err_console
            self._err_console = err_console
        return self._err_console
        # fmt: on

    @property
    def history(self) -> History:
        """Prompt history.

        Lazily instantiates the history object if it doesn't exist.
        """
        from prompt_toolkit.history import FileHistory
        from prompt_toolkit.history import InMemoryHistory

        if not self._history:
            if self.config.app.history and self.config.app.history_file:
                try:
                    self._history = FileHistory(str(self.config.app.history_file))
                except Exception as e:
                    from zabbix_cli.output.console import error

                    error(f"Failed to instantiate CLI history: {e}")
                    self._history = InMemoryHistory()
            else:
                self._history = InMemoryHistory()
        return self._history

    @property
    def ready(self) -> bool:
        """State is configured and ready to use."""
        return self.is_client_loaded and self.is_config_loaded

    def revert_config_overrides(self) -> None:
        """Revert config overrides from CLI args applied in REPL.

        Ensures that overrides only apply to a single command invocation,
        and are reset afterwards.

        In REPL mode, we have to ensure overrides don't persist between commands:

        ```
        > show_trigger_events 123 # renders table
        > -o json show_trigger_events 123 # renders json (override applied to config)
        > show_trigger_events 123 # renders table (override reverted)
        ```

        The override is reset after the command is executed.
        """
        if not self.repl or not self.is_config_loaded:
            return
        if not self._config_repl_original:
            self._config_repl_original = self.config.model_copy(deep=True)
        else:
            self.config = self._config_repl_original.model_copy(deep=True)

    def login(self) -> None:
        """Log in to the Zabbix API.

        Uses the authentication info from the config to log into the Zabbix API.

        Also sets the JSON rendering mode on the TableRenderable base class
        used to render application output.
        """
        from zabbix_cli import auth
        from zabbix_cli.models import TableRenderable

        self.client = auth.login(self.config)
        TableRenderable.legacy_json_format = self.config.app.legacy_json_format

    def logout(self) -> None:
        """Log out and clear auth token file if configured."""
        from zabbix_cli import auth

        auth.logout(self.client, self.config)

    def logout_on_exit(self) -> None:
        """Log out on exit if not keeping the session alive via auth token file."""
        # If we are NOT keeping the API session alive between CLI invocations
        # we need to remember to log out once we are done in order to end the
        # session properly.
        # https://www.zabbix.com/documentation/current/en/manual/api/reference/user/login
        if (
            # State did not finish configuration before termination
            not self.ready
            # OR We want to keep the session alive
            or self.config.app.use_session_file
            # OR we are using an API token
            or self.client.use_api_token
        ):
            return
        try:
            self.logout()
        except Exception as e:
            from zabbix_cli.exceptions import handle_exception

            # Outside of main loop, handle the exception
            handle_exception(e)


def get_state() -> State:
    """Returns the global state object.

    Instantiates a new state object with defaults if it doesn't exist.
    """
    return State()



"""DEPRECATED: Bulk execution of Zabbix commands."""

from __future__ import annotations

import subprocess
from pathlib import Path
from typing import Optional

import typer

from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import warning

app = typer.Typer(
    name="zabbix-cli-bulk-execution", help="Bulk execution of Zabbix commands"
)


@app.callback(invoke_without_command=True)
def main_callback(
    ctx: typer.Context,
    input_file: Optional[Path] = typer.Argument(
        None,
        help="File to read commands from.",
    ),
    config_file: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        help="Alternate configuration file to use.",
    ),
    input_file_legacy: Optional[Path] = typer.Option(
        None,
        "--file",
        "--input-file",
        "-f",
        hidden=True,
        help="File to read commands from.",
    ),
) -> None:
    warning(
        "zabbix-cli-bulk-execution is deprecated. Use [command]zabbix-cli --file[/] instead."
    )

    f = input_file or input_file_legacy
    if not f:
        exit_err("No input file provided. Reading from stdin is not supported.")

    # HACK: run the CLI with the file argument
    args = ["zabbix-cli", "--file", str(f)]
    if config_file:
        args.extend(["--config", str(config_file)])
    subprocess.run(args)


def main() -> None:
    """Main entry point for the CLI."""
    app()


if __name__ == "__main__":
    main()



"""REPL (Read-Eval-Print Loop) for the Zabbix CLI."""

from __future__ import annotations

import os
import shlex
import sys
from collections import defaultdict
from collections.abc import Iterable
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import NoReturn
from typing import Optional

import click
from click.exceptions import Exit as ClickExit
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.shortcuts import prompt

from zabbix_cli.exceptions import handle_exception
from zabbix_cli.output.console import err_console
from zabbix_cli.repl.completer import ClickCompleter

if TYPE_CHECKING:
    from click.core import Context

    from zabbix_cli.app import StatefulApp


class InternalCommandException(Exception):
    pass


class ExitReplException(InternalCommandException):
    pass


CommandCallable = Callable[..., Any]
"""Callable for internal commands."""


class InternalCommand(NamedTuple):
    command: CommandCallable
    description: str


_internal_commands: dict[str, InternalCommand] = {}


def _register_internal_command(
    names: Iterable[str], target: CommandCallable, description: str
):
    if isinstance(names, str):
        names = [names]

    for name in names:
        _internal_commands[name] = InternalCommand(target, description)


def _get_registered_target(
    name: str, default: Optional[CommandCallable] = None
) -> Optional[CommandCallable]:
    target_info = _internal_commands.get(name)
    if target_info:
        return target_info[0]
    return default


def _exit_internal() -> NoReturn:
    raise ExitReplException()


def _help_internal() -> str:
    formatter = click.HelpFormatter()
    formatter.write_heading("REPL help")
    formatter.indent()
    with formatter.section("External Commands"):
        formatter.write_text('prefix external commands with "!"')
    with formatter.section("Internal Commands"):
        formatter.write_text('prefix internal commands with ":"')
        info_table: defaultdict[str, list[str]] = defaultdict(list)
        for mnemonic, target_info in _internal_commands.items():
            info_table[target_info[1]].append(mnemonic)
        formatter.write_dl(
            [
                (
                    ", ".join(f":{mnemonic}" for mnemonic in sorted(mnemonics)),
                    description,
                )
                for description, mnemonics in info_table.items()
            ]
        )
    return formatter.getvalue()


_register_internal_command(["q", "quit", "exit"], _exit_internal, "exits the repl")
_register_internal_command(
    ["?", "h", "help"], _help_internal, "displays general help information"
)


def bootstrap_prompt(
    prompt_kwargs: Optional[dict[str, Any]],
    group: click.Group,
    ctx: click.Context,
    *,
    show_only_unused: bool = False,
    shortest_only: bool = False,
) -> dict[str, Any]:
    """Bootstrap prompt_toolkit kwargs or use user defined values.

    :param prompt_kwargs: The user specified prompt kwargs.
    """
    prompt_kwargs = prompt_kwargs or {}

    defaults = {
        "history": InMemoryHistory(),
        "completer": ClickCompleter(
            group, ctx, show_only_unused=show_only_unused, shortest_only=shortest_only
        ),
        "message": "> ",
    }

    for key in defaults:
        default_value = defaults[key]
        if key not in prompt_kwargs:
            prompt_kwargs[key] = default_value

    return prompt_kwargs


def register_repl(group: click.Group, name: str = "repl"):
    """Register :func:`repl()` as sub-command *name* of *group*."""
    group.command(name=name)(click.pass_context(repl))


def exit() -> None:
    """Exit the repl"""
    _exit_internal()


def dispatch_repl_commands(command: str) -> bool:
    """Execute system commands entered in the repl.

    System commands are all commands starting with "!".

    """
    if command.startswith("!") and len(command) > 1:
        os.system(command[1:])
        return True

    return False


def handle_internal_commands(command: str) -> Any:
    """Run repl-internal commands.

    Repl-internal commands are all commands starting with ":".

    """
    if command.startswith(":"):
        target = _get_registered_target(command[1:], default=None)
        if target:
            return target()


def repl(  # noqa: C901
    old_ctx: Context,
    app: StatefulApp,
    prompt_kwargs: Optional[dict[str, Any]] = None,
    *,
    allow_system_commands: bool = True,
    allow_internal_commands: bool = True,
) -> None:
    """Start an interactive shell. All subcommands are available in it.

    :param old_ctx: The current Click context.
    :param prompt_kwargs: Parameters passed to
        :py:func:`prompt_toolkit.shortcuts.prompt`.

    If stdin is not a TTY, no prompt will be printed, but only commands read
    from stdin.
    """
    # parent should be available, but we're not going to bother if not
    group_ctx = old_ctx.parent or old_ctx
    group = group_ctx.command
    isatty = sys.stdin.isatty()

    if not isinstance(group, click.Group):
        raise RuntimeError("REPL must be started from a Typer or Click group.")
    available_commands = group.commands

    # Delete the REPL command from those available
    repl_command_name = old_ctx.command.name
    if repl_command_name:
        available_commands.pop(repl_command_name, None)

    # Remove hidden commands
    available_commands = {
        cmd_name: cmd_obj
        for cmd_name, cmd_obj in available_commands.items()
        if not cmd_obj.hidden
    }

    group.commands = available_commands
    prompt_kwargs = bootstrap_prompt(prompt_kwargs, group, group_ctx)

    if isatty:

        def get_command() -> str:
            return prompt(**prompt_kwargs)

    else:
        get_command = sys.stdin.readline

    while True:
        try:
            command = get_command()
        except KeyboardInterrupt:
            continue
        except EOFError:
            break

        if not command:
            if isatty:
                continue
            else:
                break

        if allow_system_commands and dispatch_repl_commands(command):
            continue

        if allow_internal_commands:
            try:
                result = handle_internal_commands(command)
                if isinstance(result, str):
                    click.echo(result)
                    continue
            except ExitReplException:
                break

        try:
            args = shlex.split(command)
        except ValueError as e:
            click.echo(f"{type(e).__name__}: {e}")
            continue

        try:
            if app:
                group = app.as_click_group()
            with group.make_context(None, args, parent=group_ctx) as ctx:
                group.invoke(ctx)
                ctx.exit()
        except click.ClickException as e:
            e.show()
        except (ClickExit, SystemExit):
            pass
        except ExitReplException:
            break
        except Exception as e:
            try:
                # Pass exception to the exception handler
                # which handles printing and logging the exception
                # The exception handler also calls sys.exit() (used in non-interactive mode)
                # so we need to catch it and ignore it.
                handle_exception(e)
            except SystemExit:
                pass
        except KeyboardInterrupt:
            # User likely pressed Ctrl+C during a prompt or when a spinner
            # was active. Ensure message is printed on a new line.
            # TODO: determine if last char in terminal was newline somehow! Can we?
            err_console.print("\n[red]Aborted.[/]")
            pass



from __future__ import annotations

import time
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import NamedTuple
from typing import Optional
from typing import Protocol

import typer
from strenum import StrEnum

from zabbix_cli._v2_compat import ARGS_POSITIONAL
from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.config.constants import OutputFormat
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.logs import logger
from zabbix_cli.output.console import console
from zabbix_cli.output.console import err_console
from zabbix_cli.output.console import error
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import info
from zabbix_cli.output.console import success
from zabbix_cli.output.console import warning
from zabbix_cli.output.formatting.path import path_link
from zabbix_cli.output.render import render_result
from zabbix_cli.pyzabbix.enums import ExportFormat
from zabbix_cli.utils.args import parse_bool_arg
from zabbix_cli.utils.args import parse_list_arg
from zabbix_cli.utils.args import parse_path_arg
from zabbix_cli.utils.fs import open_directory
from zabbix_cli.utils.fs import sanitize_filename
from zabbix_cli.utils.utils import convert_seconds_to_duration

if TYPE_CHECKING:
    from typing_extensions import TypedDict
    from typing_extensions import Unpack

    from zabbix_cli.config.model import Config
    from zabbix_cli.pyzabbix.client import ZabbixAPI
    from zabbix_cli.pyzabbix.types import Host
    from zabbix_cli.pyzabbix.types import HostGroup
    from zabbix_cli.pyzabbix.types import Image
    from zabbix_cli.pyzabbix.types import Map
    from zabbix_cli.pyzabbix.types import MediaType
    from zabbix_cli.pyzabbix.types import Template
    from zabbix_cli.pyzabbix.types import TemplateGroup

    class ExportKwargs(TypedDict, total=False):
        hosts: list[Host]
        host_groups: list[HostGroup]
        images: list[Image]
        maps: list[Map]
        media_types: list[MediaType]
        templates: list[Template]
        template_groups: list[TemplateGroup]


HELP_PANEL = "Import/Export"


class ExportType(StrEnum):
    HOST_GROUPS = "host_groups"  # >=6.2
    TEMPLATE_GROUPS = "template_groups"  # >=6.2
    HOSTS = "hosts"
    IMAGES = "images"
    MAPS = "maps"
    TEMPLATES = "templates"
    MEDIA_TYPES = "mediaTypes"  # >= 6.0 (but should work on 5.0 too)

    @classmethod
    def _missing_(cls, value: Any) -> ExportType:
        if str(value.lower()) == "groups":
            return cls.HOST_GROUPS
        raise ValueError(f"Invalid export type: {value}")

    def human_readable(self) -> str:
        if self.value == "mediaTypes":
            return "media types"
        return self.value.replace("_", " ").lower()


class ExporterFunc(Protocol):
    def __call__(self) -> Iterator[Optional[Path]]: ...


class Exporter(NamedTuple):
    func: ExporterFunc
    type: ExportType


class ZabbixExporter:
    """Export Zabbix configuration for one or more components."""

    def __init__(
        self,
        client: ZabbixAPI,
        config: Config,
        types: list[ExportType],
        names: list[str],
        directory: Path,
        format: ExportFormat,
        timestamps: bool,
        legacy_filenames: bool,
        pretty: bool,
        ignore_errors: bool,
    ) -> None:
        self.client = client
        self.config = config
        self.export_types = types
        self.names = names
        self.directory = directory
        self.format = format
        self.timestamps = timestamps
        self.legacy_filenames = legacy_filenames
        self.pretty = pretty
        self.ignore_errors = ignore_errors

        # Ideally, we fetch and write at the same time, so we keep memory usage low,
        # while utilizing I/O and CPU as much as possible.
        # Will need to be rewritten to use threads to achieve this.

        # TODO: test that mapping contains all export types
        self.exporter_map: dict[ExportType, ExporterFunc] = {
            ExportType.HOST_GROUPS: self.export_host_groups,
            ExportType.TEMPLATE_GROUPS: self.export_template_groups,
            ExportType.HOSTS: self.export_hosts,
            ExportType.IMAGES: self.export_images,
            ExportType.MAPS: self.export_maps,
            ExportType.TEMPLATES: self.export_templates,
            ExportType.MEDIA_TYPES: self.export_media_types,
        }

        self.check_export_types()

    def run(self) -> list[Path]:
        """Run exporters."""
        files: list[Path] = []
        with err_console.status("") as status:
            for exporter in self.get_exporters():
                status.update(f"Exporting {exporter.type.human_readable()}...")
                for file in exporter.func():
                    if file:
                        files.append(file)
                success(f"Exported {exporter.type.human_readable()}")
        return files

    def check_export_types(self) -> None:
        """Check export types for compatibility."""
        # If we have no specific exports, export all object types
        for export_type in list(self.export_types):
            if not self._is_valid_export_type(export_type):
                # If the export type is not valid, remove it from the list
                self.export_types.remove(export_type)
                warning(
                    f"Export type {export_type.human_readable()} is not supported in this "
                    f"version of Zabbix ({self.client.version})."
                )

    def _is_valid_export_type(self, export_type: ExportType) -> bool:
        if export_type == ExportType.TEMPLATE_GROUPS:
            if self.client.version.release < (6, 2, 0):
                return False
        return True

    def get_exporters(self) -> list[Exporter]:
        """Get a list of exporters to run."""
        exporters: list[Exporter] = []
        for export_type in self.export_types:
            exporter = self.exporter_map.get(export_type, None)
            if not exporter:  # should never happen - tests should catch this
                raise ZabbixCLIError(
                    f"No exporter available for export type: {export_type}"
                )
            exporters.append(Exporter(exporter, export_type))
        return exporters

    def get_filename(self, name: str, id: str, export_type: ExportType) -> Path:
        """Get path to export file given a ."""
        stem = self.get_filename_stem(name, id)
        directory = self.directory / export_type.value
        return directory / f"{stem}.{self.format.value}"

    def get_filename_stem(self, name: str, id: str) -> str:
        if self.legacy_filenames:
            fn = self._get_legacy_filename_stem(name, id)
        else:
            fn = self._get_filename_stem(name, id)
        if self.timestamps:
            ts = datetime.now().strftime("%Y-%m-%dT%H%M%S%Z")
            fn = f"{fn}_{ts}"
        return sanitize_filename(fn)

    def _get_legacy_filename_stem(self, name: str, id: str) -> str:
        """Format legacy filename."""
        return f"zabbix_export_{name}_{id}"

    def _get_filename_stem(self, name: str, id: str) -> str:
        """Format filename."""
        return f"{name}_{id}"

    # TODO: refactor export methods if we want to add --ignore-errors
    # We have to find a way to generalize the export process while keeping
    # type safety intact.
    # The challenge is that we need to continue if a single export fails,
    # so we can't just wrap the call of the export method in a try/except
    # Each method needs to guard the export call of each object in a try/except
    # and that will be extremely verbose and repetitive.
    def export_host_groups(self) -> Iterator[Optional[Path]]:
        hostgroups = self.client.get_hostgroups(*self.names, search=True)
        for hg in hostgroups:
            filename = self.get_filename(hg.name, hg.groupid, ExportType.HOST_GROUPS)
            yield self.do_run_export(filename, host_groups=[hg])

    def export_template_groups(self) -> Iterator[Optional[Path]]:
        template_groups = self.client.get_templategroups(*self.names, search=True)
        for tg in template_groups:
            filename = self.get_filename(
                tg.name, tg.groupid, ExportType.TEMPLATE_GROUPS
            )
            yield self.do_run_export(filename, template_groups=[tg])

    def export_hosts(self) -> Iterator[Optional[Path]]:
        hosts = self.client.get_hosts(*self.names)
        for host in hosts:
            filename = self.get_filename(host.host, host.hostid, ExportType.HOSTS)
            yield self.do_run_export(filename, hosts=[host])

    def export_images(self) -> Iterator[Optional[Path]]:
        images = self.client.get_images(*self.names, select_image=False)
        for image in images:
            filename = self.get_filename(image.name, image.imageid, ExportType.IMAGES)
            yield self.do_run_export(filename, images=[image])

    def export_maps(self) -> Iterator[Optional[Path]]:
        maps = self.client.get_maps(*self.names)
        for m in maps:
            filename = self.get_filename(m.name, m.sysmapid, ExportType.MAPS)
            yield self.do_run_export(filename, maps=[m])

    def export_media_types(self) -> Iterator[Optional[Path]]:
        media_types = self.client.get_media_types(*self.names)
        for mt in media_types:
            filename = self.get_filename(
                mt.name, mt.mediatypeid, ExportType.MEDIA_TYPES
            )
            yield self.do_run_export(filename, media_types=[mt])

    def export_templates(self) -> Iterator[Optional[Path]]:
        templates = self.client.get_templates(*self.names)
        for template in templates:
            filename = self.get_filename(
                template.host, template.templateid, ExportType.TEMPLATES
            )
            yield self.do_run_export(filename, templates=[template])

    def do_run_export(
        self, filename: Path, **kwargs: Unpack[ExportKwargs]
    ) -> Optional[Path]:
        """Runs the export process."""
        try:
            exported = self.client.export_configuration(
                pretty=self.pretty,
                format=self.format,
                **kwargs,
            )
            return self.write_exported(exported, filename)
        except Exception as e:
            # HACKY: since we do some ugly metaprogramming to generalize the export process,
            # we don't have the actual object on hand to print a useful representation of it.
            # If every object had a __pretty__ or similar, we could use `next(iter(kwargs.values()))`
            # then get the firsty entry of that list and call __pretty__ on it. But as it stands,
            # it's prettier to just print the expected filename than to leak the entire object repr
            msg = f"Failed to export {filename}: {e}"
            if self.ignore_errors:
                error(msg, exc_info=True)
                return None
            else:
                raise ZabbixCLIError(msg) from e

    def write_exported(self, exported: str, filename: Path) -> Path:
        """Writes an exported object to a file. Returns path to file."""
        # TODO: add some logging here to show progress
        # run some callback that updates a progress bar or something
        if not filename.parent.exists():
            try:
                filename.parent.mkdir(parents=True)
            except Exception as e:
                raise ZabbixCLIError(
                    f"Failed to create directory {filename.parent}: {e}. Ensure you have permissions to create directories in the export directory."
                ) from e
        with open(filename, "w", encoding="utf-8") as f:
            f.write(exported)
        return filename


def parse_export_types(value: list[str]) -> list[ExportType]:
    # If we have no specific exports, export all object types
    if not value:
        value = list(ExportType)
    elif "#all#" in value:
        warning("#all# is a deprecated value and will be removed in a future version.")
        value = list(ExportType)
    objs: list[ExportType] = []
    for obj in value:
        try:
            export_type = ExportType(obj)
            # self._check_export_type_compat(export_type)
            objs.append(export_type)
        except ZabbixCLIError as e:
            raise e  # should this be exit_err instead?
        except Exception as e:
            raise ZabbixCLIError(f"Invalid export type: {obj}") from e
    # dedupe
    # TODO: test that StrEnum is hashable on all Python versions
    return sorted(set(objs))


def parse_export_types_callback(
    ctx: typer.Context, param: typer.CallbackParam, value: list[str]
) -> list[ExportType]:
    """Parses list of object export type names.

    In V2, this was called "objects", which isn't very descriptive...
    """
    if ctx.resilient_parsing:
        return []  # pragma: no cover
    return parse_export_types(value)


@app.command(
    name="export_configuration",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Export everything",
            "export_configuration",
        ),
        Example(
            "Export all host groups",
            "export_configuration --type host_groups",
        ),
        Example(
            "Export all host groups containing 'Linux'",
            "export_configuration --type host_groups --name '*Linux*'",
        ),
        Example(
            "Export all template groups and templates containing 'Linux' or 'Windows'",
            "export_configuration --type template_groups --type templates --name '*Linux*,*Windows*'",
        ),
    ],
)
def export_configuration(
    ctx: typer.Context,
    directory: Optional[Path] = typer.Option(
        None,
        "--directory",
        help="Directory to export configuration to. Overrides directory in config.",
        writable=True,
        file_okay=False,
    ),
    # NOTE: We can't accept comma-separated values AND multiple values when using enums!
    # Typer performs its parsing before callbacks are run, sadly.
    types: list[ExportType] = typer.Option(
        [],
        "--type",
        help="Type(s) of objects to export. Can be specified multiple times. Defaults to all object types.",
        callback=parse_export_types_callback,
    ),
    names: Optional[str] = typer.Option(
        None, "--name", help="Name(s) of objects to export. Comma-separated list."
    ),
    format: Optional[ExportFormat] = typer.Option(
        None,
        "--format",
        help="Format to export to. Overrides export format in config.",
    ),
    timestamps: Optional[bool] = typer.Option(
        None,
        "--timestamps",
        help="Add timestamps to exported filenames. Overrides config option.",
    ),
    # TODO: move/add this option to config
    legacy_filenames: bool = typer.Option(
        False,
        "--legacy-filenames",
        help="DEPRECATED: Use legacy filename scheme for exported objects.",
    ),
    pretty: bool = typer.Option(
        False,
        "--pretty",
        help="Pretty-print output. Not supported for XML.",
    ),
    open_dir: bool = typer.Option(
        False,
        "--open",
        help="Open export directory in file explorer after exporting.",
    ),
    ignore_errors: bool = typer.Option(
        False,
        "--ignore-errors",
        help="Enable best-effort exporting. Print errors but continue exporting.",
    ),
    # TODO: add --ignore-errors option
    # Legacy positional args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    r"""Export Zabbix configuration for one or more components.

    Uses defaults from Zabbix-CLI configuration file if not specified.

    [b]NOTE:[/] [option]--name[/] arguments are globs, not regex patterns.

    [b]Filename scheme is as follows:[/]

        [code]DIRECTORY/OBJECT_TYPE/NAME_ID_\[timestamp].FORMAT[/]

    [b]But it can be changed to the legacy scheme with [option]--legacy-filenames[/option]:[/b]

        [code]DIRECTORY/OBJECT_TYPE/zabbix_export_OBJECT_TYPE_NAME_ID_\[timestamp].FORMAT[/]

    Timestamps are disabled by default, but can be enabled with [option]--timestamps[/] or the [configopt]app.commands.export.timestamps[/]
    configuration option.

    Shows detailed information about exported files in JSON output mode.
    """
    from zabbix_cli.commands.results.export import ExportResult
    from zabbix_cli.models import Result

    if args:
        if not len(args) == 3:
            exit_err("Invalid number of arguments. Use options instead.")
        directory = parse_path_arg(args[0])
        types = parse_export_types(parse_list_arg(args[1]))
        names = args[2]
        # No format arg in V2...

    if legacy_filenames:
        warning(
            "--legacy-filenames is deprecated and will be removed in a future version."
        )

    # V2 compat: passing in #all# exports all names
    if names == "#all#":
        obj_names = []
    else:
        obj_names = parse_list_arg(names)

    # Options with config defaults
    exportdir = directory or app.state.config.app.commands.export.directory
    format = format or app.state.config.app.commands.export.format
    timestamps = timestamps or app.state.config.app.commands.export.timestamps

    # TODO: guard this in try/except and render useful error if it fails
    exporter = ZabbixExporter(
        client=app.state.client,
        config=app.state.config,
        types=types,
        names=obj_names,
        directory=exportdir,
        format=format,
        timestamps=timestamps,
        legacy_filenames=legacy_filenames,
        pretty=pretty,
        ignore_errors=ignore_errors,
    )
    exported = exporter.run()
    # NOTE: record duration similar to import_configuration?
    render_result(
        Result(
            message=f"Exported {len(exported)} files to {exportdir}",
            result=ExportResult(
                exported=exported, types=types, names=obj_names, format=format
            ),
            table=False,
        )
    )

    if open_dir:
        open_directory(exportdir)


class ZabbixImporter:
    def __init__(
        self,
        client: ZabbixAPI,
        config: Config,
        files: list[Path],
        create_missing: bool,
        update_existing: bool,
        delete_missing: bool,
        ignore_errors: bool,
    ) -> None:
        self.client = client
        self.config = config
        self.files = files
        self.ignore_errors = ignore_errors
        self.create_missing = create_missing
        self.update_existing = update_existing
        self.delete_missing = delete_missing

        self.imported: list[Path] = []
        self.failed: list[Path] = []

    def run(self) -> None:
        """Runs the importer."""
        from rich.progress import BarColumn
        from rich.progress import Progress
        from rich.progress import SpinnerColumn
        from rich.progress import TaskProgressColumn
        from rich.progress import TextColumn
        from rich.progress import TimeElapsedColumn

        progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeElapsedColumn(),
            transient=True,
            console=err_console,
        )
        with progress:
            task = progress.add_task("Importing files...", total=len(self.files))
            for file in self.files:
                self.import_file(file)
                progress.update(task, advance=1)

    def import_file(self, file: Path) -> None:
        # API method will return true if successful, but does failure return false
        # or does it raise an exception?
        try:
            self.client.import_configuration(file)
        except Exception as e:
            self.failed.append(file)
            msg = f"Failed to import {file}: {e}"
            if self.ignore_errors:
                error(msg, exc_info=True)
            else:
                raise ZabbixCLIError(msg) from e
        else:
            self.imported.append(file)
            logger.info(f"Imported file {file}")


def filter_valid_imports(files: list[Path]) -> list[Path]:
    """Filter list of files to include only valid imports."""
    importables = [i.casefold() for i in ExportFormat.get_importables()]
    valid: list[Path] = []
    for f in files:
        if not f.exists():
            continue
        if f.is_dir():
            continue
        if f.suffix.strip(".").casefold() not in importables:
            continue
        valid.append(f)
    return valid


@app.command(name="import_configuration", rich_help_panel=HELP_PANEL)
def import_configuration(
    ctx: typer.Context,
    to_import: Optional[str] = typer.Argument(
        None,
        help="Path to file or directory to import configuration from. Accepts glob pattern. Uses default export directory if not specified.",
    ),
    dry_run: bool = typer.Option(False, "--dryrun", help="Preview files to import."),
    create_missing: bool = typer.Option(
        True, "--create-missing/--no-create-missing", help="Create missing objects."
    ),
    update_existing: bool = typer.Option(
        True, "--update-existing/--no-update-existing", help="Update existing objects."
    ),
    delete_missing: bool = typer.Option(
        False, "--delete-missing/--no-delete-missing", help="Delete missing objects."
    ),
    ignore_errors: bool = typer.Option(
        False,
        "--ignore-errors",
        help="Enable best-effort importing. Print errors from failed imports but continue importing.",
    ),
    # Legacy positional args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Import Zabbix configuration from file, directory or glob pattern.

    Imports all files in all subdirectories if a directory is specified.
    Uses default export directory if no argument is specified.

    Determines format to import based on file extensions.
    """
    import glob

    from zabbix_cli.commands.results.export import ImportResult
    from zabbix_cli.models import Result
    from zabbix_cli.models import ReturnCode

    if args:
        if not len(args) == 2:
            exit_err("Invalid number of positional arguments. Use options instead.")
        to_import = args[0]
        dry_run = parse_bool_arg(args[1])

    # Use default export directory if no path is specified
    if not to_import:
        to_import = str(app.state.config.app.export_directory)

    # Determine if we are dealing with a directory, file or glob
    import_path = Path(to_import)
    if import_path.exists():
        if import_path.is_dir():
            files = list(import_path.glob("**/*"))
        else:
            files = [import_path]
    else:
        # Arg doesn't exist, must be glob pattern
        # If user passes in empty string, that's on them!
        files = [Path(p) for p in glob.glob(to_import)]

    files = filter_valid_imports(files)

    # HACK: in order to print a list of files without messing with line wrapping
    # and other formatting headaches, we just print using the console here
    # TODO: print properly with just render_result instead of this hack
    if dry_run:
        msg = f"Found {len(files)} files to import"
        if app.state.config.app.output.format == OutputFormat.TABLE:
            to_print = [path_link(f) for f in files]
            console.print("\n".join(to_print), highlight=False, no_wrap=True)
            info(msg)
        else:
            render_result(
                Result(
                    message=msg,
                    result=ImportResult(success=True, dryrun=True, imported=files),
                )
            )
        return

    if not files:
        exit_err(f"No files found to import matching: {to_import}")

    info(f"Found {len(files)} files to import")

    importer = ZabbixImporter(
        client=app.state.client,
        config=app.state.config,
        files=files,
        create_missing=create_missing,
        update_existing=update_existing,
        ignore_errors=ignore_errors,
        delete_missing=delete_missing,
    )

    try:
        start_time = time.monotonic()
        importer.run()
    except Exception as e:
        res = Result(
            message=f"{e}. See log for further details. Use [cyan]--ignore-errors[/] to ignore failed files.",
            return_code=ReturnCode.ERROR,
            result=ImportResult(
                success=False,
                dryrun=False,
                imported=importer.imported,
                failed=importer.failed,
            ),
            table=False,  # only render this in JSON mode
        )
    else:
        duration = time.monotonic() - start_time
        msg = f"Imported {len(importer.imported)} files in {convert_seconds_to_duration(int(duration))}"
        if importer.failed:
            msg += f", failed to import {len(importer.failed)} files"
        res = Result(
            message=msg,
            result=ImportResult(
                success=len(importer.failed) == 0,
                imported=importer.imported,
                failed=importer.failed,
                duration=duration,
            ),
            table=False,  # only render this in JSON mode
        )

    render_result(res)



"""Experimental self-update module. Powers the `update` command.

This code is largely untested and is reliant on too many external factors
to be reliably testable. It was primarily written to be used for installations
of Zabbix-CLI that are distributed as PyInstaller binaries. However, it
ended up growing in scope to encompass all the supported installation methods.

The primary use case we focus on is updating PyInstaller binaries. The other
installation methods are supported, but rely on some less-than-ideal heuristics
for detection.
"""

from __future__ import annotations

import logging
import os
import platform
import subprocess
import sys
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Any
from typing import NamedTuple
from typing import Optional

import httpx
from pydantic import BaseModel
from pydantic import ValidationError
from rich.progress import Progress
from strenum import StrEnum
from typing_extensions import Self

from zabbix_cli.__about__ import __version__
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.output.console import exit_ok
from zabbix_cli.utils.fs import make_executable
from zabbix_cli.utils.fs import move_file
from zabbix_cli.utils.fs import temp_directory

logger = logging.getLogger(__name__)

ZCLI_PACKAGES = ["zabbix-cli", "zabbix-cli-uio"]


class UpdateError(ZabbixCLIError):  # move to zabbix_cli.exceptions?
    """Base class for update errors."""


class Updater(ABC):
    """ABC for updaters."""

    def __init__(self, installation_info: InstallationInfo) -> None:
        self.installation_info = installation_info

    @abstractmethod
    def update(self) -> Optional[UpdateInfo]:
        """Update the application to the latest version.

        May return information about the update, such as the new version.
        """
        raise NotImplementedError


class GitUpdater(Updater):
    """Updater for packages installed via Git."""

    def update(self) -> UpdateInfo:
        """Update the application to the latest version."""
        # TODO: let user pass git remote, branch and other args
        subprocess.run(["git", "pull"], cwd=self.installation_info.bindir)
        # subprocess.run(["git", "-C", str(repoself.installation_info.bindir_path.absolute()), "pull"])
        return UpdateInfo(self.installation_info.method, "latest")


class PypiUpdater(Updater):
    """ABC Updater for packages installed via PyPI."""

    @property
    @abstractmethod
    def package_manager(self) -> str:
        """The name of package manager used to install the package."""
        raise NotImplementedError

    @property
    @abstractmethod
    def uninstall_command(self) -> list[str]:
        """The command used to uninstall the package."""
        raise NotImplementedError

    @property
    @abstractmethod
    def upgrade_command(self) -> list[str]:
        """The command used to upgrade the package."""
        raise NotImplementedError

    @abstractmethod
    def get_packages(self) -> set[str]:
        """Get a list of installed packages."""
        raise NotImplementedError

    def detect_package_name(self) -> str:
        """Get the name of the package to update.

        As of v3.1.3, `zabbix-cli` is released under the alias `zabbix-cli-uio` on PyPI.
        In the future, we might obtain the name `zabbix-cli` on PyPI, and
        we need to support updating both the aliased name (`zabbix-cli-uio`)
        and the actual name (`zabbix-cli`). This method detects the name of the
        package to update based on what is installed.
        """
        packages = self.get_packages()

        # BOTH zabbix-cli and zabbix-cli-uio are installed
        if "zabbix-cli-uio" in packages and "zabbix-cli" in packages:
            # BOTH zabbix-cli and zabbix-cli-uio are installed
            if self.installation_info.executable:
                if self.installation_info.executable.name in [
                    "zabbix-cli",
                    "zabbix-cli-uio",
                ]:
                    return self.installation_info.executable.name
                else:
                    raise UpdateError(
                        f"Unknown executable: {self.installation_info.executable}"
                    )
            raise UpdateError(
                "Found both [code]zabbix-cli[/] and [code]zabbix-cli-uio[/]. "
                "Unable to determine which package to update. "
                "Uninstall zabbix-cli or zabbix-cli-uio with "
                f"[code]{' '.join(self.uninstall_command)} <package>[/code] and try again."
            )
        elif "zabbix-cli-uio" in packages:
            return "zabbix-cli-uio"
        elif "zabbix-cli" in packages:
            return "zabbix-cli"
        raise UpdateError(f"Unable to detect package name from {self.package_manager}.")

    def update(self) -> None:
        """Update the package to the latest version."""
        package = self.detect_package_name()
        cmd = self.upgrade_command + [package]
        subprocess.run(cmd)


class PipxListOutput(BaseModel):
    """Model for the output of `pipx list --json`."""

    # pipx_spec_version: str # ignore this for now
    venvs: dict[str, Any]  # we just care about the keys

    @classmethod
    def from_json(cls, j: str) -> Self:
        """Parse JSON output from `pipx list --json`."""
        return cls.model_validate_json(j)

    def package_names(self) -> set[str]:
        """Get installed package names."""
        return set(self.venvs.keys())


class PipxUpdater(PypiUpdater):
    """Updater for packages installed via pipx."""

    @property
    def package_manager(self) -> str:
        """The name of package manager used to install the package."""
        return "pipx"

    @property
    def uninstall_command(self) -> list[str]:
        """The command used to uninstall the package."""
        return ["pipx", "uninstall"]

    @property
    @abstractmethod
    def upgrade_command(self) -> list[str]:
        """The command used to upgrade the package."""
        return ["pipx", "upgrade"]

    def get_packages(self) -> set[str]:
        """Get all installed pipx packages."""
        out = subprocess.check_output(["pipx", "list", "--json"], text=True)

        try:
            return PipxListOutput.from_json(out).package_names()
        except Exception as e:
            raise UpdateError(f"Unable to parse pipx list output: {e}") from e


class PipUpdater(PypiUpdater):
    """Updater for bare pip installations (PLEASE DONT USE THIS!!!!)."""

    @property
    def package_manager(self) -> str:
        """The name of package manager used to install the package."""
        return "pip"

    @property
    def uninstall_command(self) -> list[str]:
        """The command used to uninstall the package."""
        return ["pip", "uninstall"]

    @property
    @abstractmethod
    def upgrade_command(self) -> list[str]:
        """The command used to upgrade the package."""
        return ["pip", "install", "--upgrade"]

    def get_packages(self) -> set[str]:
        """Get all installed pip packages."""
        pkgs: set[str] = set()

        out = subprocess.check_output(["pip", "freeze"], text=True)
        lines = out.splitlines()
        for line in lines:
            if " @ " in line:
                pkg, _, _ = line.partition(" @ ")
            else:
                pkg, _, _ = line.partition("==")
            pkgs.add(pkg)
        return pkgs

    def detect_package_name(self) -> str:
        """Get the name of the package to update."""
        # Special case for pip, since if we are at the point where
        # we have both zabbix-cli and zabbix-cli-uio installed, we
        # might have already migrated to a mirror package on PyPI
        # where `zabbix-cli-uio` has a single dependency on `zabbix-cli`
        # in which case, we can just assume we can update `zabbix-cli`
        #
        # NOTE: there could be some edge cases where this assumption
        # is wrong, especially as we migrate to a new package name
        # however, that is such as special case - especially considering
        # NO ONE should be using pip instead of pipx or uv anyway
        packages = self.get_packages()
        if "zabbix-cli" in packages:
            return "zabbix-cli"
        elif "zabbix-cli-uio" in packages:
            return "zabbix-cli-uio"
        raise UpdateError(f"Unable to detect package name from {self.package_manager}.")


class UvUpdater(PypiUpdater):
    """Updater for packages installed via uv."""

    @property
    def package_manager(self) -> str:
        """The name of package manager used to install the package."""
        return "uv"

    @property
    def uninstall_command(self) -> list[str]:
        """The command used to uninstall the package."""
        return ["uv", "tool" "uninstall"]

    @property
    @abstractmethod
    def upgrade_command(self) -> list[str]:
        """The command used to upgrade the package."""
        return ["uv", "tool", "upgrade"]

    def get_packages(self) -> set[str]:
        """Get all installed uv packages."""
        pkgs: set[str] = set()

        out = subprocess.check_output(["uv", "tool", "list"], text=True)

        lines = out.splitlines()
        for line in lines:
            line = line.strip()
            if not line or line.startswith("-"):
                continue
            pkg, _, _ = line.partition(" ")
            pkgs.add(pkg)
        return pkgs


def download_file(url: str, destination: Path) -> None:
    """Download a file from a URL to a destination path."""
    logger.debug("Downloading %s to %s", url, destination)
    with httpx.stream("GET", url, follow_redirects=True) as response:
        with Progress() as progress:
            if response.status_code != 200:
                raise UpdateError(
                    f"Unable to download file {url}, status code {response.status_code}"
                )
            total = int(response.headers["Content-Length"])
            task = progress.add_task("[red]Downloading...[/red]", total=total)
            with open(destination, "wb") as file:
                for data in response.iter_bytes():
                    progress.update(
                        task,
                        advance=len(data),
                        description="[yellow]Downloading...[/yellow]",
                    )
                    file.write(data)
                progress.update(task, description="[green]Download Complete[/green]")
    logger.info("Downloaded %s to %s", url, destination)


class GitHubRelease(BaseModel):
    """Model for a GitHub release."""

    url: str
    tag_name: str


def get_release_arch(arch: str) -> str:
    """Get the correct arch name for a GitHub release artifact.

    Attempts to map the platform.machine() name to the name used
    in the GitHub release artifacts. If no mapping is found, the
    original name is returned.
    """
    ARCH_MAP: dict[str, str] = {
        "x86_64": "x86_64",
        "amd64": "x86_64",
        "arm64": "arm64",
        "aarch64": "arm64",
    }
    """Mapping of platform.machine() names to artifact arch names"""
    return ARCH_MAP.get(arch.lower(), arch)


def get_release_os(os: str) -> str:
    """Get the correct os name for a GitHub release artifact.

    Attempts to map the sys.platform name to the name used
    in the GitHub release artifacts. If no mapping is found, the
    original name is returned.
    """
    PLATFORM_MAP: dict[str, str] = {
        "linux": "linux",
        "darwin": "macos",
        "win32": "win",
    }
    """Mapping of sys.platform names to artifact os names"""
    return PLATFORM_MAP.get(os.lower(), os)


class PyInstallerUpdater(Updater):
    """Updater for PyInstaller installations."""

    URL_FMT = "https://github.com/unioslo/zabbix-cli/releases/latest/download/zabbix-cli-{version}-{os}-{arch}{ext}"
    """URL format for downloading the latest release."""

    URL_API_LATEST = "https://api.github.com/repos/unioslo/zabbix-cli/releases/latest"
    """URL for latest release info from API."""

    def update(self) -> Optional[UpdateInfo]:
        """Update the application to the latest version."""
        if not self.installation_info.bindir:
            raise UpdateError("Unable to determine PyInstaller binary directory")
        if not self.installation_info.executable:
            raise UpdateError("Unable to determine PyInstaller executable")

        version = self.get_release_version()
        if version == __version__:
            exit_ok(f"Application is already up-to-date ({version})")

        # The path to the binary being executed
        binfile = self.resolve_path(self.installation_info.executable)

        with temp_directory() as tmpdir:
            tmpfile = self.download(version, tmpdir)
            move_file(tmpfile, binfile)
            make_executable(binfile)

        return UpdateInfo(self.installation_info.method, version, path=binfile)

    def download(self, version: str, dest_dir: Path) -> Path:
        """Download the latest release and return the path to the downloaded file."""
        os = sys.platform
        arch = platform.machine()
        url = self.get_url(os, arch, version)
        dest = dest_dir / "zabbix-cli"
        download_file(url, dest)
        return dest

    def get_release_version(self) -> str:
        """Get the latest release info."""
        resp = httpx.get(self.URL_API_LATEST)
        if resp.status_code != 200:
            raise UpdateError(f"Failed to get latest release: {resp.text}")
        try:
            release = GitHubRelease.model_validate_json(resp.text)
        except ValidationError as e:
            raise UpdateError(
                f"Failed to parse GitHub release info: {resp.text}"
            ) from e
        if not release.tag_name:
            raise UpdateError(f"Latest GitHub release {release.url!r} has no tag name.")
        return release.tag_name

    def resolve_path(self, path: Path) -> Path:
        """Attempts to resolve a Path in strict mode, falling back to resolving aliases."""
        try:
            return path.resolve(strict=True)
        except FileNotFoundError:
            # File does not exist, assume it's a shell alias, resolve that
            alias_path = self.resolve_alias(path.name)
            # Special case for python alias - we failed to resolve it properly here
            # and ended up resolving it to a Python interpreter instead of the
            # Zabbix-CLI pyinstaller executable
            if alias_path.name == "python":
                raise UpdateError(
                    "Unable to resolve PyInstaller executable. Please update manually."
                ) from None
            return alias_path

    @staticmethod
    def resolve_alias(alias: str) -> Path:
        """Resolve a shell alias to its target path."""
        if sys.platform == "win32":
            raise UpdateError(f"Unable to resolve alias {alias!r} on Windows")

        out = subprocess.check_output(["type", "-a", alias], env=os.environ, text=True)
        lines = out.splitlines()
        logger.debug("Resolved alias %s to %s", alias, lines)
        for line in lines:
            if any(x in line for x in ["is an alias for", "is a shell function"]):
                logger.debug("Skipping line when resolving alias: %s", line)
                continue
            alias, _, path_str = line.partition(" is ")
            path_str = path_str.strip()
            return Path(path_str).resolve()
        raise UpdateError(f"Unable to resolve alias {alias}")

    # NOTE: this is a class method for ease of testing only
    @classmethod
    def get_url(cls, os: str, arch: str, version: str) -> str:
        """Get the download URL for the latest release artifact for a given platform+arch."""
        os = cls.get_release_os(os)
        arch = cls.get_release_arch(arch)
        ext = ".exe" if os == "win" else ""
        return cls.URL_FMT.format(version=version, os=os, arch=arch, ext=ext)

    @staticmethod
    def get_release_arch(arch: str) -> str:
        """Get the correct arch name for a GitHub release artifact.

        Attempts to map the platform.machine() name to the name used
        in the GitHub release artifacts. If no mapping is found, the
        original name is returned.
        """
        ARCH_MAP: dict[str, str] = {
            "x86_64": "x86_64",
            "amd64": "x86_64",
            "arm64": "arm64",
            "aarch64": "arm64",
        }
        """Mapping of platform.machine() names to artifact arch names"""
        return ARCH_MAP.get(arch.lower(), arch)

    @staticmethod
    def get_release_os(os: str) -> str:
        """Get the correct os name for a GitHub release artifact.

        Attempts to map the sys.platform name to the name used
        in the GitHub release artifacts. If no mapping is found, the
        original name is returned.
        """
        PLATFORM_MAP: dict[str, str] = {
            "linux": "linux",
            "darwin": "macos",
            "win32": "win",
        }
        """Mapping of sys.platform names to artifact os names"""
        return PLATFORM_MAP.get(os.lower(), os)


class InstallationMethod(StrEnum):
    """Installation method of the application."""

    GIT = "git"
    PIP = "pip"
    PIPX = "pipx"
    UV = "uv"
    PYINSTALLER = "pyinstaller"


class InstallationInfo(NamedTuple):
    """Information about the installation method of the application."""

    method: InstallationMethod
    package: Optional[str] = None
    executable: Optional[Path] = None
    bindir: Optional[Path] = None


def to_path(p: str) -> Optional[Path]:
    """Convert a string to a Path, returning None if the path is invalid."""
    try:
        return Path(p).expanduser().resolve()
    except Exception as e:
        logger.debug("Unable to resolve path %s: %s", p, e)
    return None


def cmd_exists(command: str, *, help: bool = True) -> bool:
    """Check if a command is available in the system."""
    cmd = [command, "--help"] if help else [command]
    try:
        subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False
    return True


# TODO: Consider moving static methods into each updater class
# or maybe even move all the detection methods into the updaters
# and have the InstallationMethodDetector just call `updater.detect()`


class InstallationMethodDetector:
    """Detects the installation method of the application."""

    def __init__(self) -> None:
        self.executable = Path(sys.argv[0])
        self.parent_dir = self.executable.parent

    @classmethod
    def detect(cls) -> InstallationInfo:
        """Attempt to detect the installation method of the application."""
        try:
            return cls()._do_detect()
        except Exception as e:
            raise UpdateError(f"Unable to detect installation method: {e}") from e

    def _do_detect(self) -> InstallationInfo:
        for method in [
            # Methods ranked in order of best->worst detection heuristics
            # as well as likelyhood of being used by the user
            self.detect_pyinstaller,
            self.detect_uv,
            self.detect_pipx,
            self.detect_pip,
            self.detect_git,
        ]:
            if (info := method()) is not None:
                logger.info("Detected installation method: %s", info.method)
                return info
        raise NotImplementedError("No detection methods succeeded")

    def detect_git(self) -> Optional[InstallationInfo]:
        """Get app installation via a local Git repository."""
        f = Path(__file__).resolve()
        package_dir = f.parent.parent
        git_dir = package_dir / ".git"
        if not git_dir.exists():
            return
        if not git_dir.is_dir():
            return
        return InstallationInfo(method=InstallationMethod.GIT, bindir=package_dir)

    @staticmethod
    def get_pipx_bin_dir() -> Optional[Path]:
        """Detect the pipx bin directory."""
        if (bd := os.environ.get("PIPX_BIN_DIR")) and (bindir := to_path(bd)):
            return bindir
        try:
            t = subprocess.check_output(
                ["pipx", "environment", "--value", "PIPX_BIN_DIR"], text=True
            )
        except subprocess.CalledProcessError as e:
            logger.error("Failed to get pipx bin dir with command %r: %s", e.cmd, e)
            return
        except FileNotFoundError:
            return  # pipx not installed
        return to_path(t)

    def detect_pipx(self) -> Optional[InstallationInfo]:
        """Detect app installation via pipx."""
        if not cmd_exists("pipx"):
            return
        bindir = self.get_pipx_bin_dir()
        if not bindir:
            return
        if not (self.parent_dir == bindir or self.parent_dir in bindir.parents):
            return
        return InstallationInfo(method=InstallationMethod.PIPX, bindir=bindir)

    @staticmethod
    def has_pip_package(pkg: str) -> bool:
        """Check if a package is installed via pip."""
        try:
            subprocess.check_call(["pip", "show", pkg])
        except subprocess.CalledProcessError:
            return False
        return True

    def detect_pip(self) -> Optional[InstallationInfo]:
        """Detect app installation via pip."""
        if not cmd_exists("pip"):
            return
        for pkg in ["zabbix-cli", "zabbix-cli-uio"]:
            if self.has_pip_package(pkg):
                return InstallationInfo(method=InstallationMethod.PIP, package=pkg)

    def detect_pyinstaller(self) -> Optional[InstallationInfo]:
        """Detect app installation via pyinstaller."""
        if not hasattr(sys, "_MEIPASS"):
            return
        # TODO: resolve alias, symlinks, etc.
        return InstallationInfo(
            method=InstallationMethod.PYINSTALLER,
            executable=self.executable,
            bindir=self.parent_dir,
        )

    @staticmethod
    def get_uv_bin_dir() -> Optional[Path]:
        """Detect the uv bin directory."""
        try:
            t = subprocess.check_output(["uv", "tool", "dir"], text=True)
        except subprocess.CalledProcessError:
            return
        except FileNotFoundError:
            return
        return to_path(t)

    @staticmethod
    def get_uv_pkg_name() -> Optional[str]:
        """Detect the installed uv package name."""
        for pkg in ZCLI_PACKAGES:
            try:
                subprocess.check_output(["uv", "tool", "list", pkg])
            except subprocess.CalledProcessError:
                continue
            return pkg

    def detect_uv(self) -> Optional[InstallationInfo]:
        """Detect app installation via uv."""
        if not cmd_exists("uv"):
            return
        bindir = self.get_uv_bin_dir()
        if not bindir:
            return
        if not (self.parent_dir == bindir or self.parent_dir in bindir.parents):
            return
        return InstallationInfo(method=InstallationMethod.UV, bindir=bindir)


UPDATERS: dict[InstallationMethod, type[Updater]] = {
    InstallationMethod.PYINSTALLER: PyInstallerUpdater,
    # InstallationMethod.GIT: GitUpdater,
    # InstallationMethod.PIP: PipUpdater,
    # InstallationMethod.PIPX: PipxUpdater,
    # InstallationMethod.UV: UvUpdater,
}


class UpdateInfo(NamedTuple):
    """Information about a completed update."""

    method: InstallationMethod
    version: str
    path: Optional[Path] = None


def get_updater(method: InstallationMethod) -> type[Updater]:
    """Get the updater class for an installation method."""
    updater = UPDATERS.get(method)
    if updater is None:
        raise UpdateError(f"No updater available for installation method {method}")
    return updater


def update(update_method: Optional[InstallationMethod] = None) -> Optional[UpdateInfo]:
    """Update the application to the latest version.

    Args:
        update_method (Optional[InstallationMethod], optional):
            Update using a specific update method. Defaults to None.
            Auto-detects the installation method if not provided.

    Raises:
        UpdateError: If the update fails.

    Returns:
        Optional[UpdateInfo]: Information about the update, such as version, method, etc.
    """
    if not update_method:
        info = InstallationMethodDetector.detect()
    else:
        info = InstallationInfo(method=update_method)

    updater = get_updater(info.method)

    try:
        return updater(info).update()
    except Exception as e:
        raise UpdateError(f"Failed to update using {info.method}: {e}") from e



from __future__ import annotations

import copy
import sys
from typing import Any
from typing import Callable

import pytest
import typer
import typer.rich_utils
from inline_snapshot import snapshot
from rich.color import Color
from rich.color import ColorType
from rich.console import Console
from rich.segment import Segment
from rich.style import Style
from typer.models import ParameterInfo
from zabbix_cli._patches import typer as typ
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.pyzabbix.enums import APIStr
from zabbix_cli.pyzabbix.enums import APIStrEnum


def test_typer_patches_idempotent() -> None:
    # patch() runs all patching functions
    typ.patch()

    # Run each patching function again to ensure no errors are raised
    # very primitive test, but ensures that we can call it as many times
    # as we want without any obvious errors.
    typ.patch__get_rich_console()
    typ.patch_generate_enum_convertor()
    typ.patch_get_click_type()
    typ.patch_help_text_spacing()
    typ.patch_help_text_style()


def test_patch__get_rich_console(
    capsys: pytest.CaptureFixture, force_color: Any
) -> None:
    original = copy.deepcopy(typer.rich_utils._get_rich_console)
    typ.patch__get_rich_console()
    new = typer.rich_utils._get_rich_console
    assert original != new

    original_val = original()
    new_val = new()
    assert original_val != new_val

    assert isinstance(original_val, Console)
    assert isinstance(new_val, Console)

    # Test some styles
    to_render = "[option]foo[/] [metavar]bar[/]"
    rendered = new_val.render(to_render)
    assert list(rendered) == snapshot(
        [
            Segment(
                text="foo",
                style=Style(
                    color=Color("cyan", ColorType.STANDARD, number=6), bold=True
                ),
            ),
            Segment(text=" ", style=Style()),
            Segment(
                text="bar",
                style=Style(
                    color=Color("yellow", ColorType.STANDARD, number=3), bold=True
                ),
            ),
            Segment(text="\n"),
        ]
    )

    # Flush capture buffer
    capsys.readouterr()
    new_val.print(to_render)
    assert capsys.readouterr().out == snapshot(
        "\x1b[1;36mfoo\x1b[0m \x1b[1;33mbar\x1b[0m\n"
    )


def _do_patch_generate_enum_convertor() -> Callable[[Any], Any]:
    original = copy.deepcopy(typer.main.generate_enum_convertor)
    typ.patch_generate_enum_convertor()
    new = typer.main.generate_enum_convertor
    assert original != new
    return new


def test_patch_generate_enum_convertor_apistrenum() -> None:
    generate_converter = _do_patch_generate_enum_convertor()

    # Test that the new convertor can handle APIStrEnum
    class APIEnum(APIStrEnum):
        FOO = APIStr("foo", 0)
        BAR = APIStr("bar", 1)

    converter = generate_converter(APIEnum)
    assert converter("foo") == APIEnum.FOO
    assert converter("bar") == APIEnum.BAR
    assert converter(0) == APIEnum.FOO
    assert converter(1) == APIEnum.BAR
    assert converter("0") == APIEnum.FOO
    assert converter("1") == APIEnum.BAR
    assert converter(APIEnum.FOO) == APIEnum.FOO
    assert converter(APIEnum.BAR) == APIEnum.BAR

    with pytest.raises(ZabbixCLIError):
        assert converter("baz")
    with pytest.raises(ZabbixCLIError):
        assert converter(2)
    with pytest.raises(ZabbixCLIError):
        assert converter("2")


def test_patch_generate_enum_convertor_strenum_lib() -> None:
    """Test patched converter with StrEnum from strenum library."""
    from strenum import StrEnum

    generate_converter = _do_patch_generate_enum_convertor()

    class TestEnum(StrEnum):
        FOO = "foo"
        BAR = "bar"

    converter = generate_converter(TestEnum)
    assert converter("foo") == TestEnum.FOO
    assert converter("bar") == TestEnum.BAR
    assert converter("FOO") == TestEnum.FOO
    assert converter("BAR") == TestEnum.BAR
    assert converter(TestEnum.FOO) == TestEnum.FOO
    assert converter(TestEnum.BAR) == TestEnum.BAR


@pytest.mark.skipif(sys.version_info < (3, 11), reason="Requires Python 3.11+")
def test_patch_generate_enum_convertor_strenum_stdlib() -> None:
    from enum import StrEnum

    generate_converter = _do_patch_generate_enum_convertor()

    class TestEnum(StrEnum):
        FOO = "foo"
        BAR = "bar"

    converter = generate_converter(TestEnum)
    assert converter("foo") == TestEnum.FOO
    assert converter("bar") == TestEnum.BAR
    assert converter("FOO") == TestEnum.FOO
    assert converter("BAR") == TestEnum.BAR
    assert converter(TestEnum.FOO) == TestEnum.FOO
    assert converter(TestEnum.BAR) == TestEnum.BAR


def test_patch_get_click_type() -> None:
    original = copy.deepcopy(typer.main.get_click_type)
    typ.patch_get_click_type()
    new = typer.main.get_click_type
    assert original != new

    class APIEnum(APIStrEnum):
        FOO = APIStr("foo", 0)
        BAR = APIStr("bar", 1)

    assert new(
        annotation=APIEnum, parameter_info=ParameterInfo()
    ).to_info_dict() == snapshot(
        {
            "param_type": "Choice",
            "name": "choice",
            "choices": ["foo", "bar", "0", "1"],
            "case_sensitive": True,
        }
    )

    assert new(
        annotation=APIEnum, parameter_info=ParameterInfo(case_sensitive=False)
    ).to_info_dict() == snapshot(
        {
            "param_type": "Choice",
            "name": "choice",
            "choices": ["foo", "bar", "0", "1"],
            "case_sensitive": False,
        }
    )


def test_patch_help_text_spacing(ctx: typer.Context) -> None:
    original = copy.deepcopy(typer.rich_utils._get_help_text)
    typ.patch_help_text_spacing()
    new = typer.rich_utils._get_help_text
    assert original != new

    ctx.command.help = "This is the first line.\n\nThis is the last line."
    help_text = new(obj=ctx.command, markup_mode="rich")
    console = typer.rich_utils._get_rich_console()
    assert list(console.render(help_text)) == snapshot(
        [
            Segment(text="This is the first line."),
            Segment(text="\n"),
            Segment(text=""),
            Segment(text="\n"),
            Segment(text="This is the last line."),
            Segment(text="\n"),
        ]
    )


def test_patch_help_text_style() -> None:
    # No in-depth testing here - just ensure that the function
    # sets the style we expect.
    # test_patch_help_text_spacing() tests the actual rendering.
    typ.patch_help_text_style()
    style = typer.rich_utils.STYLE_HELPTEXT
    assert style != "dim"
    assert style == snapshot("")



from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Optional

import typer

from zabbix_cli.config.model import Config
from zabbix_cli.config.utils import get_config


class RunMode(str, Enum):
    SHOW = "show"
    DEFAULTS = "defaults"


def main(
    arg: RunMode = typer.Argument(RunMode.DEFAULTS, case_sensitive=False),
    filename: Optional[Path] = None,
):
    """Print current or default config to stdout."""
    from zabbix_cli.logs import configure_logging

    configure_logging()
    if arg == RunMode.SHOW:
        config = get_config(filename)
    else:
        config = Config.sample_config()
    print(config.as_toml())



"""DEPRECATED: Zabbix CLI scripts.

These scripts are now baked directly into the CLI. The current scripts
are just shims to call the CLI with the appropriate arguments.
"""

from __future__ import annotations



from __future__ import annotations

import json
from typing import Any
from typing import Optional

from pytest_httpserver import HTTPServer
from werkzeug import Request
from werkzeug import Response
from zabbix_cli.pyzabbix.types import Json


def add_zabbix_endpoint(
    httpserver: HTTPServer,
    method: str,  # method is zabbix API method, not HTTP method
    *,
    params: dict[str, Any],
    response: Json,
    auth: Optional[str] = None,
    headers: Optional[dict[str, str]] = None,
    id: int = 0,
    check_id: bool = False,
) -> None:
    """Add an endpoint mocking a Zabbix API endpoint."""

    # Use a custom handler to check request contents
    def handler(request: Request) -> Response:
        # Get the JSON body of the request
        # Request has content type 'application/json-rpc'
        # so request.json() method does not work
        request_json = json.loads(request.data.decode())

        # Zabbix API method
        assert request_json["method"] == method

        # Only check the params we passed are correct
        # Missing/extra params are not checked
        for k, v in params.items():
            assert k in request_json["params"]
            assert request_json["params"][k] == v

        # Test auth token in body (< 6.4.0)
        if auth:
            assert request_json["auth"] == auth

        # Test headers
        if headers:
            for k, v in headers.items():
                assert request.headers[k] == v

        # Only check ID if we are told to
        # In parametrized tests, it can be difficult to determine the ID
        # depending on auth method, server version, etc.
        if check_id:
            assert request_json["id"] == id

        resp_json = json.dumps({"jsonrpc": "2.0", "result": response, "id": id})
        return Response(resp_json, status=200, content_type="application/json")

    httpserver.expect_oneshot_request(
        "/api_jsonrpc.php",
        method="POST",
    ).respond_with_handler(handler)


def add_zabbix_version_endpoint(
    httpserver: HTTPServer, version: str, id: int = 0
) -> None:
    """Add an endpoint emulating the Zabbix apiiinfo.version method."""
    add_zabbix_endpoint(
        httpserver,
        method="apiinfo.version",
        params={},
        response=version,
        id=id,
    )



from __future__ import annotations

import pytest
from zabbix_cli.commands.macro import fmt_macro_name
from zabbix_cli.exceptions import ZabbixCLIError

MARK_FAIL = pytest.mark.xfail(raises=ZabbixCLIError, strict=True)


@pytest.mark.parametrize(
    "name, expected",
    [
        pytest.param("my_macro", "{$MY_MACRO}"),
        pytest.param("MY_MACRO", "{$MY_MACRO}"),
        pytest.param("mY_maCrO", "{$MY_MACRO}"),
        pytest.param("foo123", "{$FOO123}"),
        pytest.param(" ", "", marks=MARK_FAIL),
        pytest.param("", "", marks=MARK_FAIL),
        pytest.param("{$}", "", marks=MARK_FAIL),
    ],
)
def test_fmt_macro_name(name: str, expected: str) -> None:
    assert fmt_macro_name(name) == expected



# Authors:
# rafael@postgresql.org.es / http://www.postgresql.org.es/
#
# Copyright (c) 2014-2016 USIT-University of Oslo
#
# This file is part of Zabbix-CLI
# https://github.com/rafaelma/zabbix-cli
#
# Zabbix-CLI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Zabbix-CLI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Zabbix-CLI.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import functools
import logging
from pathlib import Path
from typing import Any
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import overload

from pydantic import AliasChoices
from pydantic import ConfigDict
from pydantic import Field
from pydantic import PrivateAttr
from pydantic import RootModel
from pydantic import SecretStr
from pydantic import SerializationInfo
from pydantic import TypeAdapter
from pydantic import ValidationError
from pydantic import field_serializer
from pydantic import field_validator
from pydantic import model_validator
from typing_extensions import Self

from zabbix_cli._v2_compat import CONFIG_PRIORITY as CONFIG_PRIORITY_LEGACY
from zabbix_cli.bulk import BulkRunnerMode
from zabbix_cli.config.base import BaseModel
from zabbix_cli.config.commands import CommandConfig
from zabbix_cli.config.constants import AUTH_FILE
from zabbix_cli.config.constants import AUTH_TOKEN_FILE
from zabbix_cli.config.constants import HISTORY_FILE
from zabbix_cli.config.constants import LOG_FILE
from zabbix_cli.config.constants import SESSION_FILE
from zabbix_cli.config.constants import OutputFormat
from zabbix_cli.config.constants import SecretMode
from zabbix_cli.config.utils import find_config
from zabbix_cli.config.utils import fmt_deprecated_fields
from zabbix_cli.config.utils import get_deprecated_fields_set
from zabbix_cli.config.utils import load_config_conf
from zabbix_cli.config.utils import load_config_toml
from zabbix_cli.config.utils import replace_deprecated_fields
from zabbix_cli.dirs import EXPORT_DIR
from zabbix_cli.exceptions import ConfigError
from zabbix_cli.exceptions import ConfigOptionNotFound
from zabbix_cli.exceptions import PluginConfigTypeError
from zabbix_cli.exceptions import ZabbixCLIFileError
from zabbix_cli.logs import LogLevelStr
from zabbix_cli.pyzabbix.enums import ExportFormat
from zabbix_cli.utils.fs import mkdir_if_not_exists

T = TypeVar("T")

logger = logging.getLogger("zabbix_cli.config")


class APIConfig(BaseModel):
    """Configuration for the Zabbix API."""

    url: str = Field(
        default="https://zabbix.example.com",
        # Changed in V3: zabbix_api_url -> url
        validation_alias=AliasChoices("url", "zabbix_api_url"),
        description="URL of the Zabbix API host. Should not include `/api_jsonrpc.php`.",
        examples=["https://zabbix.example.com"],
    )
    username: str = Field(
        default="Admin",
        # Changed in V3: system_id -> username
        validation_alias=AliasChoices("username", "system_id"),
        description="Username for the Zabbix API.",
        examples=["Admin"],
    )
    password: SecretStr = Field(
        default=SecretStr(""),
        description="Password for user.",
        examples=["zabbix"],
    )
    auth_token: SecretStr = Field(
        default=SecretStr(""),
        description="API auth token.",
        examples=["API_TOKEN_123"],
    )
    verify_ssl: Union[bool, Path] = Field(
        default=True,
        # Changed in V3: cert_verify -> verify_ssl
        validation_alias=AliasChoices("verify_ssl", "cert_verify"),
        description="Verify SSL certificate of the Zabbix API host. Can also be a path to a CA bundle.",
    )
    timeout: Optional[int] = Field(
        default=0,
        description="API request timeout in seconds.",
    )

    @model_validator(mode="after")
    def _validate_model(self) -> Self:
        # Convert 0 timeout to None
        if self.timeout == 0:
            self.timeout = None
        return self

    @field_serializer("timeout", when_used="json")
    def _serialize_timeout(self, timeout: Optional[int]) -> int:
        """Represent None timeout as 0 in serialized output."""
        return timeout if timeout is not None else 0

    @field_serializer("password", "auth_token", when_used="json")
    def dump_secret(self, v: Any, info: SerializationInfo) -> Any:
        """Dump secrets if enabled in serialization context."""
        if not isinstance(v, SecretStr):
            logger.debug("%s from field %s is not a SecretStr", v, info)
            return v

        mode = SecretMode.from_context(info.context)
        if mode == SecretMode.PLAIN:
            return v.get_secret_value()
        elif mode == SecretMode.MASK:
            return str(v)  # SecretStr masks by default
        else:  # fall back on hidden otherwise
            return ""


class OutputConfig(BaseModel):
    """Configuration for output formatting."""

    format: OutputFormat = Field(
        default=OutputFormat.TABLE,
        description="Default output format.",
    )
    color: bool = Field(
        default=True,
        description="Use colors in terminal output.",
    )
    paging: bool = Field(
        default=False,
        description="Use paging in terminal output.",
    )
    theme: str = Field(
        default="default",
        description="Color theme to use.",
        exclude=True,
    )

    @field_validator("format", mode="before")
    @classmethod
    def _ignore_enum_case(cls, v: Any) -> Any:
        """Ignore case when validating enum value."""
        if isinstance(v, str):
            return v.lower()
        return v


class AutoUpdateConfig(BaseModel):
    """Configuration for automatic config updates."""

    enabled: bool = Field(
        default=True,
        description="Automatically update config with new config fields and migrate deprecated fields on startup.",
    )
    backup: bool = Field(
        default=True,
        description="Create a backup of the config file before updating it.",
    )


class AppConfig(BaseModel):
    """Configuration for app defaults and behavior."""

    # Auth
    use_session_file: bool = Field(
        default=True,
        validation_alias=AliasChoices("use_session_file", "use_auth_token_file"),
        description="Use session file for storing API session IDs for persistent sessions.",
    )

    session_file: Path = Field(
        default=SESSION_FILE,
        description="Path to session file.",
    )
    auth_token_file: Path = Field(
        default=AUTH_TOKEN_FILE,
        description="Path to legacy auth token file.",
        deprecated=True,
    )

    use_auth_file: bool = Field(
        default=True,
        description="Look for user credentials stored in plaintext in auth file.",
    )
    auth_file: Path = Field(
        default=AUTH_FILE,
        description="Path to auth file.",
    )
    # TODO: rename symbol to allow_insecure_auth
    allow_insecure_auth_file: bool = Field(
        default=True,
        # Changed in V3: allow_insecure_authfile -> allow_insecure_auth_file
        validation_alias=AliasChoices(
            "allow_insecure_auth",
            "allow_insecure_auth_file",
            "allow_insecure_authfile",
        ),
    )
    config_auto_update: AutoUpdateConfig = Field(
        default_factory=AutoUpdateConfig,
        description="Configuration for automatic config updates.",
    )

    # History
    history: bool = Field(
        default=True,
        description="Enable command history.",
    )
    history_file: Path = Field(
        default=HISTORY_FILE,
        description="Path to history file.",
    )

    bulk_mode: BulkRunnerMode = Field(
        default=BulkRunnerMode.STRICT,
        description="Bulk mode error handling.",
    )

    # Deprecated/moved fields
    output_format: OutputFormat = Field(
        default=OutputFormat.TABLE,
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "app.output.format"},
    )
    use_colors: bool = Field(
        default=True,
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "app.output.color"},
    )
    use_paging: bool = Field(
        default=False,
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "app.output.paging"},
    )
    system_id: str = Field(
        default="",
        validation_alias=AliasChoices("username", "system_id"),
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "api.username"},
    )
    default_hostgroups: list[str] = Field(
        default=[],
        validation_alias=AliasChoices("default_hostgroups", "default_hostgroup"),
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "app.commands.create_host.hostgroups"},
    )
    default_admin_usergroups: list[str] = Field(
        default=[],
        validation_alias=AliasChoices(
            "default_admin_usergroups", "default_admin_usergroup"
        ),
        description=(
            "Default user groups to give read/write permissions to groups "
            "created with `create_hostgroup` and `create_templategroup` "
            "when `--rw-groups` option is not provided."
        ),
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "app.commands.create_hostgroup.rw_groups"},
    )
    default_create_user_usergroups: list[str] = Field(
        default=[],
        validation_alias=AliasChoices(
            "default_create_user_usergroups", "default_create_user_usergroup"
        ),
        deprecated=True,
        exclude=True,
        json_schema_extra={
            "replacement": [
                "app.commands.create_user.usergroups",
                "app.commands.create_hostgroup.ro_groups",
            ]
        },
    )
    default_notification_users_usergroups: list[str] = Field(
        default=[],
        validation_alias=AliasChoices(
            "default_notification_users_usergroups",
            "default_notification_users_usergroup",
        ),
        deprecated=True,
        exclude=True,
        json_schema_extra={
            "replacement": "app.commands.create_notification_user.usergroups"
        },
    )
    export_directory: Path = Field(
        default=EXPORT_DIR,
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "app.commands.export.directory"},
    )
    export_format: ExportFormat = Field(
        default=ExportFormat.JSON,
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "app.commands.export.format"},
    )
    export_timestamps: bool = Field(
        default=False,
        deprecated=True,
        exclude=True,
        json_schema_extra={"replacement": "app.commands.export.timestamps"},
    )

    # Legacy options
    legacy_json_format: bool = Field(
        default=False,
        description="Use legacy JSON format.",
    )
    """Mimicks V2 behavior where the JSON output was ALWAYS a dict, where
    each entry was stored under the keys "0", "1", "2", etc.
    """

    is_legacy: bool = Field(  # TODO: use PrivateAttr instead of Field
        default=False,
        exclude=True,
    )
    """Marks whether the configuration was loaded from a legacy config file."""

    # Sub-models
    commands: CommandConfig = Field(default_factory=CommandConfig)
    output: OutputConfig = Field(default_factory=OutputConfig)

    @field_validator(
        "default_admin_usergroups",
        "default_create_user_usergroups",
        "default_hostgroups",
        "default_notification_users_usergroups",
        mode="before",
    )
    @classmethod
    def _validate_maybe_comma_separated(cls, v: Any) -> Any:
        """Validate argument that can be a single comma-separated values string
        or a list of strings.

        Used for backwards-compatibility with V2 config files, where multiple arguments
        were specified as comma-separated strings.
        """
        if isinstance(v, str):
            return v.strip().split(",")
        return v

    @field_validator("output_format", mode="before")
    @classmethod
    def _ignore_enum_case(cls, v: Any) -> Any:
        """Ignore case when validating enum value."""
        if isinstance(v, str):
            return v.lower()
        return v

    @model_validator(mode="after")
    def ensure_history_file(self) -> Self:
        if not self.history:
            return self

        if self.history_file.exists():
            if self.history_file.is_file():
                return self
            raise ConfigError(
                f"History file {self.history_file} is a directory, not a file. Disable history or specify a different path in the configuration file."
            )

        # If user passes in path to non-existent directory, we have to create that too
        # TODO: add some abstraction for creating files & directories and raising exceptions
        try:
            mkdir_if_not_exists(self.history_file.parent)
        except ZabbixCLIFileError as e:
            # TODO: print path to config file in error message
            raise ConfigError(
                f"Unable to create history file {self.history_file}. Disable history or specify a different path in the configuration file."
            ) from e
        return self


class LoggingConfig(BaseModel):
    """Configuration for application logs."""

    enabled: bool = Field(
        default=True,
        # Changed in V3: logging -> enabled (we also allow enable [why?])
        validation_alias=AliasChoices("logging", "enabled", "enable"),
        description="Enable logging.",
    )
    log_level: LogLevelStr = Field(
        default="INFO",
        description="Log level.",
    )
    log_file: Optional[Path] = Field(
        # TODO: define this default path elsewhere
        default=LOG_FILE,
        description=(
            "File for storing logs. "
            "Can be omitted to log to stderr (**warning:** NOISY)."
        ),
    )

    @field_validator("log_file", mode="before")
    @classmethod
    def _empty_string_is_none(cls, v: Any) -> Any:
        """Passing in an empty string to `log_file` sets it to `None`,
        while omitting the option altogether sets it to the default.

        Examples:
        -------
        To get `LoggingConfig.log_file == None`:

        ```toml
        [logging]
        log_file = ""
        ```

        To get `LoggingConfig.log_file == <default_log_file>`:

        ```toml
        [logging]
        # log_file = ""
        ```
        """
        if v == "":
            return None
        return v


# Can consider moving this elsewhere
@functools.cache
def _get_type_adapter(type: type[T]) -> TypeAdapter[T]:
    """Get a type adapter for a given type."""
    return TypeAdapter(type)


NotSet = object()


class PluginConfig(BaseModel):
    module: str = ""
    """Name or path to module to load.

    Should always be specified for plugins loaded from local modules.
    Can be omitted for plugins loaded from entry points.
    TOML table name is used as the module name for entry point plugins."""

    enabled: bool = True
    optional: bool = False
    """Do not raise an error if the plugin fails to load."""

    model_config = ConfigDict(extra="allow")

    # No default no type
    @overload
    def get(self, key: str) -> Any: ...

    # Type with no default
    @overload
    def get(self, key: str, *, type: type[T]) -> T: ...

    # No type with default
    @overload
    def get(self, key: str, default: T) -> Any | T: ...

    # Type with default
    @overload
    def get(
        self,
        key: str,
        default: T,
        type: type[T],
    ) -> T: ...

    # Union type with no default
    @overload
    def get(
        self,
        key: str,
        *,
        type: Optional[type[T]],
    ) -> Optional[T]: ...

    # Union type with default
    @overload
    def get(
        self,
        key: str,
        default: Optional[T],
        type: Optional[type[T]],
    ) -> Optional[T]: ...

    def get(
        self,
        key: str,
        default: Union[T, Any] = NotSet,
        type: Optional[type[T]] = object,
    ) -> Union[T, Optional[T], Any]:
        """Get a plugin configuration value by key.

        Optionally validate the value as a specific type.
        """
        try:
            if default is not NotSet:
                attr = getattr(self, key, default)
            else:
                attr = getattr(self, key)
            if type is object:
                return attr
            adapter = _get_type_adapter(type)
            return adapter.validate_python(attr)
        except AttributeError:
            raise ConfigOptionNotFound(
                f"Plugin configuration key '{key}' not found"
            ) from None
        except ValidationError as e:
            raise PluginConfigTypeError(
                f"Plugin config key '{key}' failed to validate as type {type}: {e}"
            ) from e

    def set(self, key: str, value: Any) -> None:
        """Set a plugin configuration value by key."""
        setattr(self, key, value)


class PluginsConfig(RootModel[dict[str, PluginConfig]]):
    root: dict[str, PluginConfig] = Field(default_factory=dict)

    def get(self, key: str, *, strict: bool = False) -> Optional[PluginConfig]:
        """Get a plugin configuration by name."""
        conf = self.root.get(key)
        if conf is None and strict:
            raise ConfigError(f"Plugin {key} not found in configuration")
        return conf


class Config(BaseModel):
    """Configuration for the application."""

    api: APIConfig = Field(
        default_factory=APIConfig,
        # Changed in V3: zabbix_api -> api
        validation_alias=AliasChoices("api", "zabbix_api"),
    )
    app: AppConfig = Field(
        default_factory=AppConfig,
        # Changed in V3: zabbix_config -> app
        validation_alias=AliasChoices("app", "zabbix_config"),
    )
    logging: LoggingConfig = Field(default_factory=LoggingConfig)
    plugins: PluginsConfig = Field(default_factory=PluginsConfig)

    config_path: Optional[Path] = Field(default=None, exclude=True)

    _deprecated_fields_migrated: bool = PrivateAttr(default=False)

    @property
    def sample(self) -> bool:
        # No fields set means this is a sample config
        return not bool(self.model_fields_set)

    @model_validator(mode="after")
    def _set_deprecated_fields_in_new_location(self) -> Self:
        """Set values specified on deprecated fields in their new location
        and optionally update the config file on disk.

        I.e. `app.username` -> `api.username`.
             `app.output_format` -> `app.output.format`.

        Only updates new fields if the old field is set and the new field is not.
        """
        if self._deprecated_fields_migrated:
            return self

        # Guard against failure in case we have a config that cannot be updated
        # After all, this validator is added for _increased_ compatibility, not decreased
        try:
            replace_deprecated_fields(self)
            self.check_emit_upgrade_instructions()
            if self.app.config_auto_update.enabled:
                self.dump_updated_config()
        except Exception as e:
            logger.error("Failed to update deprecated fields in config: %s", e)

        # Mark as migrated even if we failed to avoid repeated attempts
        self._deprecated_fields_migrated = True
        return self

    @classmethod
    def sample_config(cls) -> Config:
        """Get a sample configuration."""
        return cls()

    @classmethod
    def from_file(
        cls, filename: Optional[Path] = None, *, init: bool = False
    ) -> Config:
        """Load configuration from a file.

        Attempts to find a config file to load if none is specified.

        Prioritizes V3 .toml config files, but falls back
        to legacy V2 .conf config files if no .toml config file can be found.
        """
        if filename:
            fp = filename
        else:
            fp = find_config(filename)
            if (
                not fp
            ):  # We didn't find a .toml file, so try to find a legacy .conf file
                fp = find_config(filename, CONFIG_PRIORITY_LEGACY)

        # Failed to find both .toml and .conf files
        if not fp or not fp.exists():
            if init:
                from zabbix_cli.config.utils import init_config

                return init_config(config_file=filename)
            else:
                return cls.sample_config()

        if fp.suffix == ".conf":
            return cls.from_conf_file(fp)
        else:
            return cls.from_toml_file(fp)

    @classmethod
    def from_toml_file(cls, filename: Path) -> Config:
        """Load configuration from a TOML file."""
        conf = load_config_toml(filename)
        try:
            return cls(**conf, config_path=filename)
        except ValidationError as e:
            raise ConfigError(f"Invalid configuration file {filename}: {e}") from e
        except Exception as e:
            raise ConfigError(
                f"Failed to load configuration file {filename}: {e}"
            ) from e

    @classmethod
    def from_conf_file(cls, filename: Path) -> Config:
        """Load configuration from a legacy .conf file."""
        logging.info("Using legacy config file (%s)", filename)
        conf = load_config_conf(filename)
        # Use legacy JSON format if we load from a legacy .conf file
        # and mark the loaded config as stemming from a legacy config file
        conf.setdefault("zabbix_config", {}).setdefault("legacy_json_format", True)
        conf.setdefault("zabbix_config", {}).setdefault("is_legacy", True)
        try:
            return cls(**conf, config_path=filename)
        except ValidationError as e:
            raise ConfigError(
                f"Failed to validate legacy configuration file {filename}: {e}"
            ) from e
        except Exception as e:
            raise ConfigError(
                f"Failed to load legacy configuration file {filename}: {e}"
            ) from e

    def as_toml(self, secrets: SecretMode = SecretMode.MASK) -> str:
        """Dump the configuration to a TOML string."""
        import tomli_w

        try:
            return tomli_w.dumps(
                self.model_dump(
                    mode="json",
                    exclude_none=True,  # we shouldn't have any, but just in case
                    context={"secrets": secrets},
                )
            )
        except Exception as e:
            raise ConfigError(f"Failed to serialize configuration to TOML: {e}") from e

    def dump_to_file(
        self, filename: Path, secrets: SecretMode = SecretMode.HIDE
    ) -> None:
        """Dump the configuration to a TOML file."""
        try:
            mkdir_if_not_exists(filename.parent)
            filename.write_text(self.as_toml(secrets=secrets))
        except OSError as e:
            raise ConfigError(
                f"Failed to write configuration file {filename}: {e}"
            ) from e

    def check_emit_upgrade_instructions(self) -> None:
        from zabbix_cli.output.console import warning

        # TODO: Detmermine whether or not to prefix command with `zabbix-cli`
        #       based on whether or not we are in the REPL

        # Legacy configs provide different instructions
        # and we cannot use the same heuristics for determining deprecations
        if self.app.is_legacy:
            warning(
                "Your configuration file is from an older version of Zabbix-CLI.\n"
                "  To update your config file to the new format, run:\n"
                "  [command]zabbix-cli migrate_config[/]\n"
                "  For more information, see the documentation."
            )
            return

        # Auto-updates do not require user intervention
        if self.app.config_auto_update.enabled:
            return

        deprecated_fields = get_deprecated_fields_set(self)
        if not deprecated_fields:
            return

        # Deprecated fields have no replacements – nothing to inform about
        # NOTE: unless we want to provide a way to remove deprecated fields?
        if all(not field.replacement for field in deprecated_fields):
            return

        warning(
            "Your configuration file contains deprecated options.\n"
            "  To update your config file with the new options, run:\n"
            "  [command]zabbix-cli update_config[/]\n"
            "  For more information, see the documentation."
        )

    def dump_updated_config(self, config_file: Optional[Path] = None) -> None:
        """Update deprecated fields in the configuration to their new location.

        Uses the current config values to update the config file on disk.
        """
        from zabbix_cli.output.console import info
        from zabbix_cli.output.console import success
        from zabbix_cli.output.formatting.path import path_link

        deprecated_fields = get_deprecated_fields_set(self)
        if not deprecated_fields:
            return  # no deprecated fields
        elif not any(field.replacement for field in deprecated_fields):
            return  # no replacements for deprecated fields

        config_file = config_file or self.config_path
        if not config_file:
            logger.warning("Cannot update configuration file: no config file path set.")
            return

        if self.app.config_auto_update.backup:
            backup_file = config_file.with_suffix(config_file.suffix + ".bak")
            try:
                backup_file.write_text(config_file.read_text())
            except OSError as e:
                raise ConfigError(
                    f"Failed to create backup of config file {config_file} "
                    f"before updating it: {e}. "
                    "Disable backups with [option]app.config_auto_update.backup=false[/] "
                    "to proceed without a backup."
                ) from e
            logger.info("Created backup of config file at %s", backup_file)

        try:
            self.dump_to_file(config_file, secrets=SecretMode.PLAIN)
        except ConfigError as e:
            raise ConfigError(
                f"Failed to update config file {config_file} with new fields: {e}. "
                "Disable automatic config updates with [option]app.config_auto_update.enabled=false[/] "
                "to proceed without automatic updates."
            ) from e

        success(f"Updated config file {path_link(config_file)}.")
        info(f"Deprecated fields updated:\n{fmt_deprecated_fields(deprecated_fields)}")



from __future__ import annotations

from collections.abc import Mapping
from enum import Enum
from typing import Any
from typing import Generic
from typing import Optional
from typing import TypeVar

from strenum import StrEnum
from typing_extensions import Self

from zabbix_cli.exceptions import ZabbixCLIError

T = TypeVar("T")


class APIStr(str, Generic[T]):
    """String type that can be used as an Enum choice while also
    carrying an API value associated with the string.
    """

    # Instance variables are set by __new__
    api_value: T
    value: str
    metadata: Mapping[str, Any]
    hidden: bool

    def __new__(
        cls,
        s: str,
        api_value: T = None,
        metadata: Optional[Mapping[str, Any]] = None,
        *,
        hidden: bool = False,
    ) -> APIStr[T]:
        if isinstance(s, APIStr):
            return s  # type: ignore # Type checker should be able to infer generic type
        if api_value is None:
            raise ZabbixCLIError("API value must be provided for APIStr.")
        obj = str.__new__(cls, s)
        obj.value = s
        obj.api_value = api_value
        obj.metadata = metadata or {}
        obj.hidden = hidden
        return obj


MixinType = TypeVar("MixinType", bound="Choice")


class Choice(Enum):
    """Enum subclass that allows for an Enum to have APIStr values, which
    enables it to be instantiated with either the name of the option
    or the Zabbix API value of the option.

    We can instantiate the enum with either the name or the API value:
        * `ActiveInterface("available")`
        * `ActiveInterface(1)`
        * `ActiveInterface("1")`

    We use these enums as choices in the CLI, so that users can pass in
    a human readable name for the choice or its API value.

    Since the API itself is inconsistent with usage of strings and ints,
    we support instantiation with either one.

    Provides the `from_prompt` class method, which prompts the user to select
    one of the enum members. The prompt text is generated from the class name
    by default, but can be overridden by setting the `__choice_name__` class var.

    Also provides a method for returning the API value of an enum member with the
    with the `as_api_value()` method.
    """

    value: APIStr[int]  # pyright: ignore[reportIncompatibleMethodOverride]
    __choice_name__: str = ""  # default (falls back to class name)

    def __new__(cls, value: APIStr[int]) -> Choice:
        # Adds type checking for members in enum definition
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def __str__(self) -> str:
        return str(self.value)

    def casefold(self) -> str:
        return str(self.value).casefold()

    @classmethod
    def __fmt_name__(cls) -> str:
        """Return the name of the enum class in a human-readable format.

        If no default is provided, the class name is split on capital letters and
        lowercased, e.g. `ActiveInterface` becomes `active interface`.
        """
        if cls.__choice_name__:
            return cls.__choice_name__
        return (
            "".join([(" " + i if i.isupper() else i) for i in cls.__name__])
            .lower()
            .strip()
        )

    # NOTE: should we use a custom prompt class instead of repurposing the str prompt?
    @classmethod
    def from_prompt(
        cls: type[MixinType],
        prompt: Optional[str] = None,
        default: Optional[MixinType] = None,
    ) -> MixinType:
        """Prompt the user to select a choice from the enum.

        Args:
            prompt (Optional[str], optional): Alternative prompt.
                Defaults to None, which uses the formatted class name.

        Returns:
            MixinType: Enum member selected by the user.
        """
        from zabbix_cli.output.prompts import str_prompt

        if not prompt:
            # Assume
            prompt = cls.__fmt_name__()
            # Uppercase first letter without mangling the rest of the string
            if prompt and prompt[0].islower():
                prompt = prompt[0].upper() + prompt[1:]

        default_val = str(default) if default else None

        choice = str_prompt(
            prompt,
            choices=cls.choices(),
            default=default_val,
        )
        return cls(choice)

    @classmethod
    def public_members(cls) -> list[Self]:
        """Return list of visible enum members."""
        return [e for e in cls if not e.value.hidden]

    @classmethod
    def choices(cls) -> list[str]:
        """Return list of string values of the enum members."""
        return [str(e) for e in cls.public_members()]

    @classmethod
    def api_choices(cls) -> list[int]:
        """Return list of API values of the enum members."""
        # NOTE: should this be a list of strings instead?
        return [e.as_api_value() for e in cls.public_members()]

    @classmethod
    def all_choices(cls) -> list[str]:
        """All public choices as well as their API values."""
        return cls.choices() + [str(c) for c in cls.api_choices()]

    def as_api_value(self) -> int:
        """Return the equivalent Zabbix API value."""
        return self.value.api_value

    @classmethod
    def _missing_(cls, value: object) -> object:
        """Method that is called when an enum member is not found.

        Attempts to find the member with 2 strategies:
        1. Search for a member with the given string value (ignoring case)
        2. Search for a member with the given API value (converted to string)
        """
        for v in cls:
            if v.value == value:
                return v
            # kinda hacky. Should make sure we are dealing with strings here:
            elif str(v.value).lower() == str(value).lower():
                return v
            elif str(v.as_api_value()) == str(value):
                return v
        raise ZabbixCLIError(f"Invalid {cls.__fmt_name__()}: {value!r}.")


class APIStrEnum(Choice):
    """Enum that returns value of member as str."""

    # FIXME: should inherit from string somehow!
    # Does not inherit from str now, as it would convert enum member value
    # to string, thereby losing the API associated value.
    # If we are to do that, we need to hijack the object creation and inject
    # the member value somehow?
    def as_status(self, default: str = "Unknown", *, with_code: bool = False) -> str:
        return self.string_from_value(self.value, default=default, with_code=with_code)

    @classmethod
    def string_from_value(
        cls: type[Self],
        value: Any,
        default: str = "Unknown",
        *,
        with_code: bool = False,
    ) -> str:
        """Get a formatted status string given a value."""
        try:
            c = cls(value)
            # All lowercase is capitalized
            if str(c.value.islower()):
                name = c.value.capitalize()
            # Everything else is left as is
            else:
                name = str(c.value)
            code = c.value.api_value
        except (ValueError, ZabbixCLIError):
            name = default
            code = value
        if with_code:
            return f"{name} ({code})"
        return name


class AckStatus(APIStrEnum):
    NO = APIStr("no", 0)
    YES = APIStr("yes", 1)


class ActiveInterface(APIStrEnum):
    """Active interface availability status."""

    __choice_name__ = "Agent availability status"

    UNKNOWN = APIStr("unknown", 0)
    AVAILABLE = APIStr("available", 1)
    UNAVAILABLE = APIStr("unavailable", 2)


class DataCollectionMode(APIStrEnum):
    """Maintenance data collection mode."""

    ON = APIStr("on", 0)
    OFF = APIStr("off", 1)


class EventStatus(APIStrEnum):
    OK = APIStr("OK", 0)
    PROBLEM = APIStr("PROBLEM", 1)


class ExportFormat(StrEnum):
    XML = "xml"
    JSON = "json"
    YAML = "yaml"
    PHP = "php"

    @classmethod
    def _missing_(cls, value: object) -> ExportFormat:
        """Case-insensitive missing lookup.

        Allows for both `ExportFormat("JSON")` and `ExportFormat("json")`, etc.
        """
        if not isinstance(value, str):
            raise TypeError(f"Invalid format: {value!r}. Must be a string.")
        value = value.lower()
        for e in cls:
            if e.value.lower() == value:
                return e
        raise ValueError(f"Invalid format: {value!r}.")

    @classmethod
    def get_importables(cls) -> list[ExportFormat]:
        """Return list of formats that can be imported."""
        return [cls.JSON, cls.YAML, cls.XML]


class GUIAccess(APIStrEnum):
    """GUI Access for a user group."""

    __choice_name__ = "GUI Access"

    DEFAULT = APIStr("default", 0)
    INTERNAL = APIStr("internal", 1)
    LDAP = APIStr("ldap", 2)
    DISABLE = APIStr("disable", 3)


class HostgroupFlag(APIStrEnum):
    """Hostgroup flags."""

    PLAIN = APIStr("plain", 0)
    DISCOVER = APIStr("discovered", 4)


class HostgroupType(APIStrEnum):
    """Hostgroup types."""

    NOT_INTERNAL = APIStr("not internal", 0)
    INTERNAL = APIStr("internal", 1)


class InterfaceConnectionMode(APIStrEnum):
    """Interface connection mode.

    Controls the value of `useip` when creating interfaces in the API.
    """

    DNS = APIStr("DNS", 0)
    IP = APIStr("IP", 1)


class InterfaceType(APIStrEnum):
    """Interface type."""

    AGENT = APIStr("Agent", 1, metadata={"port": "10050"})
    SNMP = APIStr("SNMP", 2, metadata={"port": "161"})
    IPMI = APIStr("IPMI", 3, metadata={"port": "623"})
    JMX = APIStr("JMX", 4, metadata={"port": "12345"})

    def get_port(self: InterfaceType) -> str:
        """Returns the default port for the given interface type."""
        try:
            return self.value.metadata["port"]
        except KeyError:
            raise ZabbixCLIError(f"Unknown interface type: {self}") from None


class InventoryMode(APIStrEnum):
    """Host inventory mode."""

    DISABLED = APIStr("disabled", -1)
    MANUAL = APIStr("manual", 0)
    AUTOMATIC = APIStr("automatic", 1)


class ItemType(APIStrEnum):
    ZABBIX_AGENT = APIStr("Zabbix agent", 0)
    SNMPV1_AGENT = APIStr("SNMPv1 agent", 1)
    ZABBIX_TRAPPER = APIStr("Zabbix trapper", 2)
    SIMPLE_CHECK = APIStr("Simple check", 3)
    SNMPV2_AGENT = APIStr("SNMPv2 agent", 4)
    ZABBIX_INTERNAL = APIStr("Zabbix internal", 5)
    SNMPV3_AGENT = APIStr("SNMPv3 agent", 6)
    ZABBIX_AGENT_ACTIVE = APIStr("Zabbix agent (active)", 7)
    ZABBIX_AGGREGATE = APIStr("Zabbix aggregate", 8)
    WEB_ITEM = APIStr("Web item", 9)
    EXTERNAL_CHECK = APIStr("External check", 10)
    DATABASE_MONITOR = APIStr("Database monitor", 11)
    IPMI_AGENT = APIStr("IPMI agent", 12)
    SSH_AGENT = APIStr("SSH agent", 13)
    TELNET_AGENT = APIStr("TELNET agent", 14)
    CALCULATED = APIStr("calculated", 15)
    JMX_AGENT = APIStr("JMX agent", 16)
    SNMP_TRAP = APIStr("SNMP trap", 17)
    DEPENDENT_ITEM = APIStr("Dependent item", 18)
    HTTP_AGENT = APIStr("HTTP agent", 19)
    SNMP_AGENT = APIStr("SNMP agent", 20)
    SCRIPT = APIStr("Script", 21)


class MacroType(APIStrEnum):
    TEXT = APIStr("text", 0)
    SECRET = APIStr("secret", 1)
    VAULT_SECRET = APIStr("vault secret", 2)


class MacroAutomatic(APIStrEnum):
    """Macro automatic (discovered) status."""

    NO = APIStr("no", 0)  # user managed
    YES = APIStr("yes", 1)  # managed by discovery rule


class MaintenancePeriodType(APIStrEnum):
    """Maintenance period."""

    ONETIME = APIStr("one time", 0)
    DAILY = APIStr("daily", 2)
    WEEKLY = APIStr("weekly", 3)
    MONTHLY = APIStr("monthly", 4)


class MaintenanceStatus(APIStrEnum):
    """Host maintenance status."""

    # API values are inverted here compared to monitoring status...
    ON = APIStr("on", 1)
    OFF = APIStr("off", 0)


class MaintenanceType(APIStrEnum):
    """Maintenance type."""

    WITH_DC = APIStr("With DC", 0)
    WITHOUT_DC = APIStr("Without DC", 1)


class MaintenanceWeekType(APIStrEnum):
    """Maintenance every week type."""

    FIRST_WEEK = APIStr("first week", 1)
    SECOND_WEEK = APIStr("second week", 2)
    THIRD_WEEK = APIStr("third week", 3)
    FOURTH_WEEK = APIStr("fourth week", 4)
    LAST_WEEK = APIStr("last week", 5)


class MediaTypeType(APIStrEnum):
    """Type of media type."""

    EMAIL = APIStr("email", 0)
    SCRIPT = APIStr("script", 1)
    SMS = APIStr("SMS", 2)
    WEBHOOK = APIStr("webhook", 4)
    # no 3 value in API


class MonitoredBy(APIStrEnum):  # >=7.0 only
    SERVER = APIStr("server", 0)
    PROXY = APIStr("proxy", 1)
    PROXY_GROUP = APIStr("proxygroup", 2)


class MonitoringStatus(APIStrEnum):
    """Host monitoring status."""

    ON = APIStr("on", 0)  # Yes, 0 is on, 1 is off...
    OFF = APIStr("off", 1)
    UNKNOWN = APIStr(
        "unknown", 3, hidden=True
    )  # Undocumented, but shows up in virtual trigger hosts (get_triggers(select_hosts=True))


class ProxyCompatibility(APIStrEnum):
    """Proxy compatibility status for >=7.0"""

    UNDEFINED = APIStr("undefined", 0)
    CURRENT = APIStr("current", 1)
    OUTDATED = APIStr("outdated", 2)
    UNSUPPORTED = APIStr("unsupported", 3)


class ProxyGroupState(APIStrEnum):
    UNKNOWN = APIStr("unknown", 0)
    OFFLINE = APIStr("offline", 1)
    RECOVERING = APIStr("recovering", 2)
    ONLINE = APIStr("online", 3)
    DEGRADING = APIStr("degrading", 4)


class ProxyMode(APIStrEnum):
    """Proxy mode."""

    ACTIVE = APIStr("active", 0)
    PASSIVE = APIStr("passive", 1)


class ProxyModePre70(APIStrEnum):
    """Proxy mode pre 7.0."""

    ACTIVE = APIStr("active", 5)
    PASSIVE = APIStr("passive", 6)


class SNMPAuthProtocol(APIStrEnum):
    """Authentication protocol for SNMPv3."""

    __choice_name__ = "SNMPv3 auth protocol"

    MD5 = APIStr("MD5", 0)
    SHA1 = APIStr("SHA1", 1)
    # >=6.0 only:
    SHA224 = APIStr("SHA224", 2)
    SHA256 = APIStr("SHA256", 3)
    SHA384 = APIStr("SHA384", 4)
    SHA512 = APIStr("SHA512", 5)


class SNMPPrivProtocol(APIStrEnum):
    """Privacy protocol for SNMPv3."""

    __choice_name__ = "SNMPv3 privacy protocol"

    DES = APIStr("DES", 0)
    AES = APIStr("AES", 1)  # < 6.0 only
    # >=6.0 only:
    AES128 = APIStr("AES128", 1)  # >= 6.0
    AES192 = APIStr("AES192", 2)
    AES256 = APIStr("AES256", 3)
    AES192C = APIStr("AES192C", 4)
    AES256C = APIStr("AES256C", 5)


class SNMPSecurityLevel(APIStrEnum):
    __choice_name__ = "SNMPv3 security level"

    # Match casing from Zabbix API
    NO_AUTH_NO_PRIV = APIStr("noAuthNoPriv", 0)
    AUTH_NO_PRIV = APIStr("authNoPriv", 1)
    AUTH_PRIV = APIStr("authPriv", 2)


class TriggerPriority(APIStrEnum):
    UNCLASSIFIED = APIStr("unclassified", 0)
    INFORMATION = APIStr("information", 1)
    WARNING = APIStr("warning", 2)
    AVERAGE = APIStr("average", 3)
    HIGH = APIStr("high", 4)
    DISASTER = APIStr("disaster", 5)


class UsergroupPermission(APIStrEnum):
    """Usergroup permission levels."""

    DENY = APIStr("deny", 0)
    READ_ONLY = APIStr("ro", 2)
    READ_WRITE = APIStr("rw", 3)


class UsergroupStatus(APIStrEnum):
    """Usergroup status."""

    ENABLED = APIStr("enabled", 0)
    DISABLED = APIStr("disabled", 1)


class UserRole(APIStrEnum):
    __choice_name__ = "User role"

    # Match casing from Zabbix API
    USER = APIStr("user", 1)
    ADMIN = APIStr("admin", 2)
    SUPERADMIN = APIStr("superadmin", 3)
    GUEST = APIStr("guest", 4)


class ValueType(APIStrEnum):
    NUMERIC_FLOAT = APIStr("Numeric (float)", 0)
    CHARACTER = APIStr("Character", 1)
    LOG = APIStr("Log", 2)
    NUMERIC_UNSIGNED = APIStr("Numeric (unsigned)", 3)
    TEXT = APIStr("Text", 4)



from __future__ import annotations

from typing import Literal
from typing import Union

from zabbix_cli.models import TableRenderable
from zabbix_cli.pyzabbix.types import Host
from zabbix_cli.pyzabbix.types import HostGroup
from zabbix_cli.pyzabbix.types import Template
from zabbix_cli.pyzabbix.types import TemplateGroup


class LinkTemplateToHostResult(TableRenderable):
    host: str
    templates: list[str]
    action: str

    @classmethod
    def from_result(
        cls,
        templates: list[Template],
        host: Host,
        action: str,
    ) -> LinkTemplateToHostResult:
        to_link: set[str] = set()  # names of templates to link
        for t in templates:
            for h in t.hosts:
                if h.host == host.host:
                    break
            else:
                to_link.add(t.host)
        return cls(
            host=host.host,
            templates=sorted(to_link),
            action=action,
        )


class UnlinkTemplateFromHostResult(TableRenderable):
    host: str
    templates: list[str]
    action: str

    @classmethod
    def from_result(
        cls,
        templates: list[Template],
        host: Host,
        action: str,
    ) -> UnlinkTemplateFromHostResult:
        """Only show templates that are actually unlinked."""
        to_remove: set[str] = set()
        for t in templates:
            for h in t.hosts:
                if h.host == host.host:
                    to_remove.add(t.host)  # name of template
                    break
        return cls(
            host=host.host,
            templates=list(to_remove),
            action=action,
        )


class LinkTemplateResult(TableRenderable):
    """Result type for (un)linking templates to templates."""

    source: list[str]
    destination: list[str]
    action: str

    @classmethod
    def from_result(
        cls,
        source: list[Template],
        destination: list[Template],
        action: Literal["Link", "Unlink", "Unlink and clear"],
    ) -> LinkTemplateResult:
        return cls(
            source=[t.host for t in source],
            destination=[t.host for t in destination],
            action=action,
        )


class TemplateGroupResult(TableRenderable):
    templates: list[str]
    groups: list[str]

    @classmethod
    def from_result(
        cls,
        templates: list[Template],
        groups: Union[list[TemplateGroup], list[HostGroup]],
    ) -> TemplateGroupResult:
        return cls(
            templates=[t.host for t in templates],
            groups=[h.name for h in groups],
        )


class RemoveTemplateFromGroupResult(TableRenderable):
    group: str
    templates: list[str]

    @classmethod
    def from_result(
        cls,
        templates: list[Template],
        group: Union[TemplateGroup, HostGroup],
    ) -> RemoveTemplateFromGroupResult:
        to_remove: set[str] = set()
        for template in group.templates:
            for t in templates:
                if t.host == template.host:
                    to_remove.add(t.host)
                    break
        return cls(
            templates=list(to_remove),
            group=group.name,
        )



from __future__ import annotations

import importlib
from pathlib import Path


def bootstrap_commands() -> None:
    """Bootstrap all command defined in the command modules."""
    module_dir = Path(__file__).parent
    for module in module_dir.glob("*.py"):
        if module.stem == "__init__":
            continue
        importlib.import_module(f".{module.stem}", package=__package__)



from __future__ import annotations

import pytest
from inline_snapshot import snapshot
from zabbix_cli.exceptions import ZabbixAPIException
from zabbix_cli.exceptions import ZabbixAPIRequestError
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.exceptions import get_cause_args
from zabbix_cli.pyzabbix.types import ZabbixAPIError
from zabbix_cli.pyzabbix.types import ZabbixAPIResponse


@pytest.mark.parametrize(
    "outer_t", [TypeError, ValueError, ZabbixCLIError, ZabbixAPIException]
)
def test_get_cause_args(outer_t: type[Exception]) -> None:
    try:
        try:
            try:
                raise ZabbixAPIException("foo!")
            except ZabbixAPIException as e:
                raise TypeError("foo", "bar") from e
        except TypeError as e:
            raise outer_t("outer") from e
    except outer_t as e:
        args = get_cause_args(e)
        assert args == snapshot(["outer", "foo", "bar", "foo!"])


def test_get_cause_args_no_cause() -> None:
    e = ZabbixAPIException("foo!")
    args = get_cause_args(e)
    assert args == snapshot(["foo!"])


def test_get_cause_args_with_api_response() -> None:
    api_resp = ZabbixAPIResponse(
        jsonrpc="2.0",
        result=None,
        id=1,
        error=ZabbixAPIError(code=-123, message="Some error"),
    )
    e = ZabbixAPIRequestError("foo!", api_response=api_resp)
    args = get_cause_args(e)
    assert args == snapshot(["foo!", "(-123) Some error"])


def test_get_cause_args_with_api_response_with_data() -> None:
    """Get the cause args from an exception with an API response with data."""
    api_resp = ZabbixAPIResponse(
        jsonrpc="2.0",
        result=None,
        id=1,
        error=ZabbixAPIError(code=-123, message="Some error", data='{"foo": 42}'),
    )
    e = ZabbixAPIRequestError("foo!", api_response=api_resp)
    args = get_cause_args(e)
    assert args == snapshot(["foo!", '(-123) Some error {"foo": 42}'])



from __future__ import annotations

from datetime import datetime

import pytest
from zabbix_cli.pyzabbix.enums import ProxyGroupState
from zabbix_cli.pyzabbix.types import CreateHostInterfaceDetails
from zabbix_cli.pyzabbix.types import DictModel
from zabbix_cli.pyzabbix.types import Event
from zabbix_cli.pyzabbix.types import GlobalMacro
from zabbix_cli.pyzabbix.types import Host
from zabbix_cli.pyzabbix.types import HostGroup
from zabbix_cli.pyzabbix.types import HostInterface
from zabbix_cli.pyzabbix.types import Image
from zabbix_cli.pyzabbix.types import ImportRules
from zabbix_cli.pyzabbix.types import Item
from zabbix_cli.pyzabbix.types import Macro
from zabbix_cli.pyzabbix.types import MacroBase
from zabbix_cli.pyzabbix.types import Maintenance
from zabbix_cli.pyzabbix.types import Map
from zabbix_cli.pyzabbix.types import MediaType
from zabbix_cli.pyzabbix.types import ProblemTag
from zabbix_cli.pyzabbix.types import Proxy
from zabbix_cli.pyzabbix.types import ProxyGroup
from zabbix_cli.pyzabbix.types import Role
from zabbix_cli.pyzabbix.types import Template
from zabbix_cli.pyzabbix.types import TemplateGroup
from zabbix_cli.pyzabbix.types import TimePeriod
from zabbix_cli.pyzabbix.types import Trigger
from zabbix_cli.pyzabbix.types import UpdateHostInterfaceDetails
from zabbix_cli.pyzabbix.types import User
from zabbix_cli.pyzabbix.types import Usergroup
from zabbix_cli.pyzabbix.types import UserMedia
from zabbix_cli.pyzabbix.types import ZabbixAPIBaseModel
from zabbix_cli.pyzabbix.types import ZabbixRight


@pytest.mark.parametrize(
    "model",
    [
        # TODO: replace with hypothesis tests when Pydantic v2 support is available
        # Test in order of definition in types.py
        ZabbixRight(permission=2, id="str"),
        User(userid="123", username="test"),
        Usergroup(name="test", usrgrpid="123", gui_access=0, users_status=0),
        Template(templateid="123", host="test"),
        TemplateGroup(groupid="123", name="test", uuid="test123"),
        HostGroup(name="test", groupid="123"),
        DictModel(),
        Host(hostid="123"),
        HostInterface(
            type=1, main=1, ip="127.0.0.1", dns="", port="10050", useip=1, bulk=1
        ),
        CreateHostInterfaceDetails(version=2),
        UpdateHostInterfaceDetails(),
        Proxy(proxyid="123", name="test", address="127.0.0.1"),
        ProxyGroup(
            proxy_groupid="123",
            name="test",
            description="yeah",
            failover_delay="60",
            min_online="1",
            state=ProxyGroupState.ONLINE,
        ),
        MacroBase(macro="foo", value="bar", type=0, description="baz"),
        Macro(
            hostid="123",
            hostmacroid="1234",
            macro="foo",
            value="bar",
            type=0,
            description="baz",
        ),
        GlobalMacro(
            globalmacroid="123g", macro="foo", value="bar", type=0, description="baz"
        ),
        Item(itemid="123"),
        Role(roleid="123", name="test", type=1, readonly=0),
        MediaType(mediatypeid="123", name="test", type=0),
        UserMedia(mediatypeid="123", sendto="foo@example.com"),
        TimePeriod(period=123, timeperiod_type=2),
        ProblemTag(tag="foo", operator=2, value="bar"),
        Maintenance(maintenanceid="123", name="test"),
        Event(
            eventid="source",
            object=1,
            objectid="123",
            source=2,
            acknowledged=0,
            clock=datetime.now(),
            name="test",
            severity=2,
        ),
        Trigger(triggerid="123"),
        Image(imageid="123", name="test", imagetype=1),
        Map(sysmapid="123", name="test", width=100, height=100),
        ImportRules.get(),
    ],
)
def test_model_dump(model: ZabbixAPIBaseModel) -> None:
    """Test that the model can be dumped to JSON."""
    try:
        model.model_dump_json()
    except Exception as e:
        pytest.fail(f"Failed to dump model {model} to JSON: {e}")

    try:
        model.model_dump_api()
    except Exception as e:
        pytest.fail(f"Failed to dump model {model} to API-compatible dict: {e}")

    try:
        model.model_dump()
    except Exception as e:
        pytest.fail(f"Failed to dump model {model} to dict: {e}")



from __future__ import annotations

from collections.abc import Generator
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import pytest
import typer
from packaging.version import Version
from typer.testing import CliRunner
from zabbix_cli.app import StatefulApp
from zabbix_cli.config.model import Config
from zabbix_cli.main import app
from zabbix_cli.pyzabbix.client import ZabbixAPI
from zabbix_cli.state import State
from zabbix_cli.state import get_state

runner = CliRunner()


@pytest.fixture(name="app")
def _app() -> Iterator[StatefulApp]:
    yield app


@pytest.fixture
def ctx(app: StatefulApp) -> typer.Context:
    """Create context for the main command."""
    # Use the CliRunner to invoke a command and capture the context
    obj = {}
    with runner.isolated_filesystem():

        @app.callback(invoke_without_command=True)
        def callback(ctx: typer.Context):
            obj["ctx"] = ctx  # Capture the context in a non-local object

        runner.invoke(app, [], obj=obj)
    return obj["ctx"]


DATA_DIR = Path(__file__).parent / "data"

# Read sample configs once per test run to avoid too much I/O
TOML_CONFIG = DATA_DIR / "zabbix-cli.toml"
TOML_CONFIG_STR = TOML_CONFIG.read_text()

CONF_CONFIG = DATA_DIR / "zabbix-cli.conf"
CONF_CONFIG_STR = CONF_CONFIG.read_text()


@pytest.fixture()
def data_dir() -> Iterator[Path]:
    yield Path(__file__).parent / "data"


@pytest.fixture()
def config_path(tmp_path: Path) -> Iterator[Path]:
    config_copy = tmp_path / "zabbix-cli.toml"
    config_copy.write_text(TOML_CONFIG_STR)
    yield config_copy


@pytest.fixture()
def legacy_config_path(tmp_path: Path) -> Iterator[Path]:
    config_copy = tmp_path / "zabbix-cli.conf"
    config_copy.write_text(CONF_CONFIG_STR)
    yield config_copy


@pytest.fixture(name="state")
def state(config: Config, zabbix_client: ZabbixAPI) -> Iterator[State]:
    """Return a fresh State object with a config and client.

    The client is not logged in to the Zabbix API.

    Modifies the State singleton to ensure a fresh state is returned
    each time.
    """
    State._instance = None  # pyright: ignore[reportPrivateUsage]
    state = get_state()
    state.config = config
    state.client = zabbix_client
    yield state
    # reset after test
    State._instance = None  # pyright: ignore[reportPrivateUsage]


@pytest.fixture(name="config")
def config(tmp_path: Path) -> Iterator[Config]:
    """Return a sample config."""
    conf = Config.sample_config()
    # Set up logging for the test environment
    log_file = tmp_path / "zabbix-cli.log"
    conf.logging.log_file = log_file
    conf.logging.log_level = "DEBUG"  # we want to see all logs
    yield conf


@pytest.fixture(name="zabbix_client")
def zabbix_client() -> Iterator[ZabbixAPI]:
    config = Config.sample_config()
    client = ZabbixAPI.from_config(config)
    yield client


@pytest.fixture(name="zabbix_client_mock_version")
def zabbix_client_mock_version(
    zabbix_client: ZabbixAPI, monkeypatch: pytest.MonkeyPatch
) -> Iterator[ZabbixAPI]:
    monkeypatch.setattr(zabbix_client, "api_version", lambda: Version("7.0.0"))
    yield zabbix_client


@pytest.fixture(name="force_color")
def force_color() -> Generator[Any, Any, Any]:
    import os

    os.environ["FORCE_COLOR"] = "1"
    yield
    os.environ.pop("FORCE_COLOR", None)


@pytest.fixture(name="no_color")
def no_color() -> Generator[Any, Any, Any]:
    """Disable color in a test. Takes precedence over force_color."""
    import os

    os.environ.pop("FORCE_COLOR", None)
    os.environ["NO_COLOR"] = "1"
    yield
    os.environ.pop("NO_COLOR", None)



from __future__ import annotations

import logging
import math
import os
from functools import cache
from functools import wraps
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Optional
from typing import overload

from rich.prompt import Confirm
from rich.prompt import FloatPrompt
from rich.prompt import IntPrompt
from rich.prompt import Prompt
from typing_extensions import ParamSpec
from typing_extensions import TypeVar

from zabbix_cli.exceptions import ZabbixCLIError

from .console import err_console
from .console import error
from .console import exit_err
from .formatting.path import path_link
from .style import Color
from .style import Icon
from .style import green

T = TypeVar("T")
P = ParamSpec("P")


def no_headless(f: Callable[P, T]) -> Callable[P, T]:
    """Decorator that causes application to exit if called from a headless environment
    when the prompt does not have a default value (i.e. required input).

    If the prompt has a default value, that value is returned instead."""

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        if is_headless():
            # If a default argument was passed in, we can return that:
            default = kwargs.get("default")
            if "default" in kwargs and default is not ...:
                logging.debug(
                    "Returning default value from %s(%s, %s)", f, args, kwargs
                )
                return default  # type: ignore # bit of a hack
            # Assume first arg is the prompt:
            prompt = args[0] if args else kwargs.get("prompt", "")
            exit_err(
                f"Cannot proceed! User input required for {prompt!r}. Exiting...",
                args=args,
                kwargs=kwargs,
            )
        return f(*args, **kwargs)

    return wrapper


HEADLESS_VARS_SET = ["CI", "ZABBIX_CLI_HEADLESS"]
"""Envvars that indicate headless environ when set (1, true)."""
HEADLESS_VARS_SET_MAP = {"DEBIAN_FRONTEND": "noninteractive"}
"""Envvars that indicate headless environ when set to a specific value."""

TRUE_ARGS = ["1", "true"]  # case-insensitive


# NOTE: if testing this function, clear cache after each test
@cache
def is_headless() -> bool:
    """Determines if we are running in a headless environment (e.g. CI, Docker, etc.)"""
    # Truthy values indicate headless
    for envvar in HEADLESS_VARS_SET:
        if os.environ.get(envvar, "").lower() in ["1", "true"]:
            return True

    # Specific values indicate headless
    for envvar, value in HEADLESS_VARS_SET_MAP.items():
        if os.environ.get(envvar, None) == value:
            return True

    import shellingham  # pyright: ignore[reportMissingTypeStubs]

    try:
        shellingham.detect_shell()  # pyright: ignore[reportUnknownMemberType]
    except shellingham.ShellDetectionFailure:
        # If we can't detect the shell, assume we're in a headless environment
        return True

    return False


def prompt_msg(*msgs: str) -> str:
    return f"[bold]{green(Icon.PROMPT)} {' '.join(msg.strip() for msg in filter(None, msgs))}[/bold]"


@no_headless
def str_prompt(
    prompt: str,
    *,
    default: str | None = None,  # pyright: ignore[reportArgumentType] # rich uses ... to signify no default
    password: bool = False,
    choices: Optional[list[str]] = None,
    show_default: bool = True,
    show_choices: bool = True,
    empty_ok: bool = False,
    strip: bool = True,
) -> str:
    """Prompts the user for a string input.

    Optionally controls for empty input. Loops until a valid input is provided.

    Args:
        prompt (str): String prompt to display to the user.
        default (Any, optional): Default value to use if the user does not provide input.
            If not provided, the user will be required to provide input.
        password (bool, optional): Whether to hide the input. Defaults to False.
        choices (list[str], optional): List of valid choices. If provided, the user
            must select one of the choices. Defaults to None.
        show_default (bool, optional): Show the default value. Defaults to True.
            If `password=True`, this option is overridden and set to False.
        show_choices (bool, optional): Show the choices in the prompt. Defaults to True.
        empty_ok (bool, optional): Allow input consisting of no characters or only whitespace.
            Defaults to False.
        strip (bool, optional): Strip whitespace from the input. Defaults to True.
            Must be `False` to preserve whitespace when `empty_ok=True`.

    Returns:
        str: The user input as a string. If `strip=True`, leading and trailing whitespace
            will be removed from the input.
    """

    prompt_parts = [prompt]

    # Never show password default or empty default
    if password or not default:
        show_default = False

    # Add notice for hidden password default
    if password and default:
        prompt_parts.append("(leave empty to use existing value)")

    msg = prompt_msg(*prompt_parts)

    # Rich uses `...` to signify no default, which is clunky when wrapping
    # Prompt.ask, since it doesn't have an overload for explicitly passing
    # in `...`. To work around that, we don't pass in the default kwarg
    # if it's None.
    kwargs: dict[str, Any] = {}
    if default is not None:
        kwargs["default"] = default

    while not (
        inp := Prompt.ask(
            msg,
            console=err_console,
            password=password,
            show_default=show_default,
            show_choices=show_choices,
            choices=choices,
            **kwargs,
        )
    ):
        if empty_ok:  # nothing else to check
            break

        if not inp:
            error("Input cannot be empty.")
        elif inp.isspace() and inp != default:
            error("Input cannot solely consist of whitespace.")
        else:
            break
    return inp.strip() if strip else inp


@no_headless
def str_prompt_optional(
    prompt: str,
    *,
    default: str = "",
    password: bool = False,
    show_default: bool = False,
    choices: Optional[list[str]] = None,
    strip: bool = True,
) -> str:
    prompt = f"{prompt} [i](optional)[/]"
    return str_prompt(
        prompt,
        default=default,
        password=password,
        show_default=show_default,
        choices=choices,
        empty_ok=True,
        strip=strip,
    )


TypeConstructor = Callable[[object], T]


@no_headless
def list_prompt(
    prompt: str,
    *,
    default: list[T] | None = None,
    empty_ok: bool = True,
    strip: bool = True,
    keep_empty: bool = False,
    # https://github.com/python/mypy/issues/3737
    # https://github.com/python/mypy/issues/3737#issuecomment-1446769973
    # Using this weird TypeConstructor type seems very hacky
    type: TypeConstructor[T] = str,
) -> list[T]:
    """Prompt user for a comma-separated list of values."""
    from zabbix_cli.utils.args import parse_list_arg

    default_arg = ",".join(str(d) for d in default) if default else None

    inp = str_prompt(
        prompt,
        default=default_arg,
        empty_ok=empty_ok,
        strip=strip,
    )
    arglist = parse_list_arg(inp, keep_empty=keep_empty)
    try:
        # NOTE: type() in this context is the constructor for the type
        # we passed in, not the built-in type() function (shadowed in this scope)
        return [type(arg) for arg in arglist]
    except Exception as e:
        raise ZabbixCLIError(f"Invalid value: {e}") from e


@no_headless
def int_prompt(
    prompt: str,
    *,
    default: int | None = None,
    show_default: bool = True,
    min: int | None = None,
    max: int | None = None,
    show_range: bool = True,
    **kwargs: Any,
) -> int:
    """Prompt user for an integer input. Loops until a valid input is provided.

    Args:
        prompt (str): String prompt to display to the user.
        default (int, optional): Default value to use if the user does not provide input.
            If not provided, the user will be required to provide input.
        show_default (bool, optional): Show the default value. Defaults to True.
        min (int, optional): Minimum value. Defaults to None.
        max (int, optional): Maximum value. Defaults to None.
        show_range (bool, optional): Show the range in the prompt. Defaults to True.
        **kwargs: Additional keyword arguments to pass to the prompt.

    Returns:
        int: The user input as an integer.
    """
    return _number_prompt(
        IntPrompt,
        prompt,
        default=default,
        show_default=show_default,
        min=min,
        max=max,
        show_range=show_range,
        **kwargs,
    )


@no_headless
def float_prompt(
    prompt: str,
    *,
    default: float | None = None,
    show_default: bool = True,
    min: float | None = None,
    max: float | None = None,
    show_range: bool = True,
    **kwargs: Any,
) -> float:
    """Prompt user for a float input. Loops until a valid input is provided.

    Args:
        prompt (str): String prompt to display to the user.
        default (float, optional): Default value to use if the user does not provide input.
            If not provided, the user will be required to provide input.
        show_default (bool, optional): Show the default value. Defaults to True.
        min (float, optional): Minimum value. Defaults to None.
        max (float, optional): Maximum value. Defaults to None.
        show_range (bool, optional): Show the range in the prompt. Defaults to True.
        **kwargs: Additional keyword arguments to pass to the prompt.

    Returns:
        float: The user input as a float.
    """
    val = _number_prompt(
        FloatPrompt,
        prompt,
        default=default,
        show_default=show_default,
        min=min,
        max=max,
        show_range=show_range,
    )
    # explicit cast to float since users might pass in int as default
    # and we have no logic inside _number_prompt to handle that
    return float(val)


@overload
def _number_prompt(
    prompt_type: type[IntPrompt],
    prompt: str,
    *,
    default: int | float | None = ...,
    show_default: bool = ...,
    min: int | float | None = ...,
    max: int | float | None = ...,
    show_range: bool = ...,
) -> int: ...


@overload
def _number_prompt(
    prompt_type: type[FloatPrompt],
    prompt: str,
    *,
    default: int | float | None = ...,
    show_default: bool = ...,
    min: int | float | None = ...,
    max: int | float | None = ...,
    show_range: bool = ...,
) -> float: ...


def _number_prompt(
    prompt_type: type[IntPrompt] | type[FloatPrompt],
    prompt: str,
    *,
    default: int | float | None = None,
    show_default: bool = True,
    min: int | float | None = None,
    max: int | float | None = None,
    show_range: bool = True,
) -> int | float:
    prompt_parts = [prompt]
    if show_range:
        input_range_str = ""
        if min is not None and max is not None:
            if min > max:
                raise ValueError("min must be less than or equal to max")
            input_range_str = f"{min}<=x<={max}"
        elif min is not None:
            input_range_str = f"x>={min}"
        elif max is not None:
            input_range_str = f"x<={max}"
        if input_range_str:
            prompt_parts.append(Color.YELLOW(input_range_str))

    msg = prompt_msg(*prompt_parts)

    # See str_prompt() for why we use pass default as part of the kwargs mapping
    kwargs: dict[str, Any] = {}
    if default is not None:
        kwargs["default"] = default

    while True:
        val = prompt_type.ask(  # pyright: ignore[reportUnknownVariableType]
            msg,
            console=err_console,
            show_default=show_default,
            **kwargs,
        )

        # Shouldn't happen, but ask() returns DefaultType | int | float
        # so it thinks we could have an ellipsis here
        if not isinstance(val, (int, float)):
            error("Value must be a number")
            continue
        if math.isnan(val):
            error("Value can't be NaN")
            continue
        if min is not None and val < min:
            error(f"Value must be greater or equal to {min}")
            continue
        if max is not None and val > max:
            error(f"Value must be less than or equal to {max}")
            continue
        return val


@no_headless
def bool_prompt(
    prompt: str,
    *,
    default: bool | None = None,
    show_default: bool = True,
) -> bool:
    """Prompt user for a boolean input (y/n). Loops until a valid input is provided.

    Args:
        prompt (str): String prompt to display to the user.
        default (bool, optional): Default value to use if the user does not provide input.
            If not provided, the user will be required to provide input.
        show_default (bool, optional): Show the default value. Defaults to True.
        warning (bool, optional): Show a warning message if the user selects "no". Defaults to False.
        **kwargs: Additional keyword arguments to pass to the prompt.
    Returns:
        bool: The user input as a boolean.
    """
    kwargs: dict[str, Any] = {}
    if default is not None:
        kwargs["default"] = default

    return Confirm.ask(
        prompt_msg(prompt),
        console=err_console,
        show_default=show_default,
        **kwargs,
    )


@no_headless
def path_prompt(
    prompt: str,
    default: str | Path | None = None,
    *,
    show_default: bool = True,
    exist_ok: bool = True,
    must_exist: bool = False,
) -> Path:
    """Prompt user for a path.

    Optionally checks if the path exists OR if it does _not_ exist.

    Args:
        prompt (str): String prompt to display to the user.
        default (str | Path, optional): Default value to use if the user does not provide input.
            If not provided, the user will be required to provide input.
        show_default (bool, optional): Show the default value. Defaults to True.
        exist_ok (bool, optional): Allow existing paths. Defaults to True.
        must_exist (bool, optional): Require existing paths. Defaults to False.

    Returns:
        The user input as a Path object.
    """
    if isinstance(default, Path):
        default_arg = str(default)
    else:
        default_arg = default

    while True:
        path_str = str_prompt(
            prompt,
            default=default_arg,
            show_default=show_default,
        )
        path = Path(path_str)

        if must_exist and not path.exists():
            error(f"Path does not exist: {path_link(path)}")
        elif not exist_ok and path.exists():
            error(f"Path already exists: {path_link(path)}")
        else:
            return path



from __future__ import annotations

import logging
from typing import Any

import pytest
from inline_snapshot import snapshot
from zabbix_cli.config.constants import OutputFormat
from zabbix_cli.exceptions import ZabbixAPIException
from zabbix_cli.exceptions import ZabbixAPIRequestError
from zabbix_cli.output.console import RESERVED_EXTRA_KEYS
from zabbix_cli.output.console import debug
from zabbix_cli.output.console import debug_kv
from zabbix_cli.output.console import error
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import get_extra_dict
from zabbix_cli.output.console import info
from zabbix_cli.output.console import success
from zabbix_cli.output.console import warning
from zabbix_cli.pyzabbix.types import ZabbixAPIError
from zabbix_cli.pyzabbix.types import ZabbixAPIResponse
from zabbix_cli.state import State


@pytest.mark.parametrize(
    "inp,expect",
    [
        # No extras
        ({}, {}),
        # 1 extra
        ({"name": "foo"}, {"name_": "foo"}),
        # 2 extras
        ({"name": "foo", "level": "DEBUG"}, {"name_": "foo", "level_": "DEBUG"}),
        # 3 extras (2 reserved)
        (
            {"name": "foo", "level": "DEBUG", "key": "value"},
            {"name_": "foo", "level_": "DEBUG", "key": "value"},
        ),
    ],
)
def test_get_extra_dict(inp: dict[str, Any], expect: dict[str, Any]) -> None:
    extra = get_extra_dict(**inp)
    assert extra == expect


def test_get_extra_dict_reserved_keys() -> None:
    """Test that all reserved keys are renamed."""
    d: dict[str, Any] = {}
    for key in RESERVED_EXTRA_KEYS:
        d[key] = key
    extra = get_extra_dict(**d)
    assert extra == snapshot(
        {
            "name_": "name",
            "level_": "level",
            "pathname_": "pathname",
            "lineno_": "lineno",
            "msg_": "msg",
            "args_": "args",
            "exc_info_": "exc_info",
            "func_": "func",
            "sinfo_": "sinfo",
        }
    )


def test_debug_kv(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture, state: State
) -> None:
    state.config.logging.log_level = "DEBUG"
    caplog.set_level(logging.DEBUG)
    debug_kv("some", "error")
    captured = capsys.readouterr()
    assert captured.err == snapshot("some                : error\n")
    assert caplog.record_tuples == snapshot(
        [("zabbix_cli", 10, "some                : error")]
    )
    assert caplog.records[0].funcName == snapshot("test_debug_kv")


def test_debug(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture, state: State
) -> None:
    state.config.logging.log_level = "DEBUG"
    caplog.set_level(logging.DEBUG)
    debug("Some error")
    captured = capsys.readouterr()
    assert captured.err == snapshot("Some error\n")
    assert caplog.record_tuples == snapshot([("zabbix_cli", 10, "Some error")])
    assert caplog.records[0].funcName == snapshot("test_debug")


def test_info(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture
) -> None:
    caplog.set_level(logging.INFO)
    info("Some error")
    captured = capsys.readouterr()
    assert captured.err == snapshot("! Some error\n")
    assert caplog.record_tuples == snapshot([("zabbix_cli", 20, "Some error")])
    assert caplog.records[0].funcName == snapshot("test_info")


def test_success(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture
) -> None:
    caplog.set_level(logging.INFO)
    success("Some error")
    captured = capsys.readouterr()
    assert captured.err == snapshot("✓ Some error\n")
    assert caplog.record_tuples == snapshot([("zabbix_cli", 20, "Some error")])
    assert caplog.records[0].funcName == snapshot("test_success")


def test_warning(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture
) -> None:
    caplog.set_level(logging.INFO)
    warning("Some error")
    captured = capsys.readouterr()
    assert captured.err == snapshot("⚠ Some error\n")
    assert caplog.record_tuples == snapshot([("zabbix_cli", 30, "Some error")])
    assert caplog.records[0].funcName == snapshot("test_warning")


def test_error(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture
) -> None:
    caplog.set_level(logging.INFO)
    error("Some error")
    captured = capsys.readouterr()
    assert captured.err == snapshot("✗ ERROR: Some error\n")
    assert caplog.record_tuples == snapshot([("zabbix_cli", 40, "Some error")])
    assert caplog.records[0].funcName == snapshot("test_error")


def test_exit_err_table(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture, state: State
) -> None:
    state.config.app.output.format = OutputFormat.TABLE
    caplog.set_level(logging.INFO)
    with pytest.raises(SystemExit):
        exit_err("Some error")
    captured = capsys.readouterr()
    assert captured.err == snapshot("✗ ERROR: Some error\n")
    assert captured.out == snapshot("")


def test_exit_err_json(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture, state: State
) -> None:
    state.config.app.output.format = OutputFormat.JSON
    with pytest.raises(SystemExit):
        exit_err("Some error")
    captured = capsys.readouterr()
    assert captured.err == snapshot("✗ ERROR: Some error\n")
    assert captured.out == snapshot(
        """\
{
  "message": "Some error",
  "errors": [],
  "return_code": "Error",
  "result": null
}
"""
    )
    assert caplog.record_tuples == snapshot([("zabbix_cli", 40, "Some error")])


def test_exit_err_json_with_errors(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture, state: State
) -> None:
    outer_exc = TypeError("Outer exception")
    outer_exc.__cause__ = ValueError("Inner exception")

    state.config.app.output.format = OutputFormat.JSON
    with pytest.raises(SystemExit):
        exit_err("Some error", exception=outer_exc)
    captured = capsys.readouterr()
    assert captured.err == snapshot("✗ ERROR: Some error\n")
    assert captured.out == snapshot(
        """\
{
  "message": "Some error",
  "errors": [
    "Outer exception",
    "Inner exception"
  ],
  "return_code": "Error",
  "result": null
}
"""
    )
    assert caplog.record_tuples == snapshot([("zabbix_cli", 40, "Some error")])


def test_exit_err_json_with_zabbix_api_request_error(
    capsys: pytest.CaptureFixture[str], caplog: pytest.LogCaptureFixture, state: State
) -> None:
    api_resp = ZabbixAPIResponse(
        jsonrpc="2.0",
        result=None,
        id=1,
        error=ZabbixAPIError(
            code=-123, message="API response error", data='{"foo": 42}'
        ),
    )
    outer_exc = ZabbixAPIException("Outer exception")
    outer_exc.__cause__ = ZabbixAPIRequestError(
        "Inner exception", api_response=api_resp
    )

    state.config.app.output.format = OutputFormat.JSON
    with pytest.raises(SystemExit):
        exit_err("Some error", exception=outer_exc)
    captured = capsys.readouterr()
    assert captured.err == snapshot("✗ ERROR: Some error\n")
    assert captured.out == snapshot(
        """\
{
  "message": "Some error",
  "errors": [
    "Outer exception",
    "Inner exception",
    "(-123) API response error {\\"foo\\": 42}"
  ],
  "return_code": "Error",
  "result": null
}
"""
    )
    assert caplog.record_tuples == snapshot([("zabbix_cli", 40, "Some error")])



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel
from pydantic import ConfigDict

from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.pyzabbix.enums import ActiveInterface
from zabbix_cli.pyzabbix.enums import MaintenanceStatus
from zabbix_cli.pyzabbix.enums import MonitoringStatus


# TODO: don't use BaseModel for this
# Use a normal class with __init__ instead
class HostFilterArgs(BaseModel):
    """Unified processing of old filter string and new filter options."""

    active: Optional[ActiveInterface] = None
    maintenance_status: Optional[MaintenanceStatus] = None
    status: Optional[MonitoringStatus] = None

    model_config = ConfigDict(validate_assignment=True)

    @classmethod
    def from_command_args(
        cls,
        filter_legacy: Optional[str],
        active: Optional[ActiveInterface],
        maintenance: Optional[bool],
        monitored: Optional[bool],
    ) -> HostFilterArgs:
        args = cls()
        if filter_legacy:
            items = filter_legacy.split(",")
            for item in items:
                try:
                    key, value = (s.strip("'\"") for s in item.split(":"))
                except ValueError as e:
                    raise ZabbixCLIError(
                        f"Failed to parse filter argument at: {item!r}"
                    ) from e
                if key == "available":
                    args.active = ActiveInterface(value)
                elif key == "maintenance":
                    args.maintenance_status = MaintenanceStatus(value)
                elif key == "status":
                    args.status = MonitoringStatus(value)
        else:
            if active is not None:
                args.active = active
            if monitored is not None:
                # Inverted API values (0 = ON, 1 = OFF) - use enums directly
                args.status = MonitoringStatus.ON if monitored else MonitoringStatus.OFF
            if maintenance is not None:
                args.maintenance_status = (
                    MaintenanceStatus.ON if maintenance else MaintenanceStatus.OFF
                )
        return args
